from pathlib import Path

import numpy as np
import numpy.testing as np_test
import pytest
import scipy.io

from sunpy.time import parse_time

from irispy import utils
from irispy.data.test import ROOTDIR

# Arrays for the fit_iris_xput method
raw_response_data = scipy.io.readsav(Path(ROOTDIR) / ".." / "iris_sra_c_20161022.geny", python_dict=True, verbose=False)
iris_response = {name: raw_response_data["p0"][name][0] for name in raw_response_data["p0"].dtype.names}
time_obs = [parse_time("2013-09-03")]
time_cal_coeffs0 = parse_time(iris_response.get("C_F_TIME"), format="utime")
time_cal_coeffs1 = parse_time(iris_response.get("C_N_TIME"), format="utime")
cal_coeffs0 = iris_response.get("COEFFS_FUV")[0, :, :]
cal_coeffs1 = iris_response.get("COEFFS_FUV")[1, :, :]
cal_coeffs2 = iris_response.get("COEFFS_FUV")[2, :, :]
cal_coeffs3 = iris_response.get("COEFFS_NUV")[0, :, :]
cal_coeffs4 = iris_response.get("COEFFS_NUV")[1, :, :]
cal_coeffs5 = iris_response.get("COEFFS_NUV")[2, :, :]
cal_coeffs6 = iris_response.get("COEFFS_NUV")[3, :, :]
iris_fit_expected0 = np.array([0.79865301])
iris_fit_expected1 = np.array([2.2495413])
iris_fit_expected2 = np.array([2.2495413])
iris_fit_expected3 = np.array([0.23529011])
iris_fit_expected4 = np.array([0.25203046])
iris_fit_expected5 = np.array([0.25265095])
iris_fit_expected6 = np.array([0.253833])


def test_get_iris_response_greater_than_6():
    with pytest.raises(KeyError):
        utils.get_iris_response(time_obs, response_version=13)


# Tests for get_iris_response function
# Version 1
sav_file_path1 = Path(ROOTDIR) / "idl_iris_get_response_20130903_new_version001.sav"
test_iris_response1 = scipy.io.readsav(sav_file_path1, python_dict=True, verbose=False)
iris_response_load1 = test_iris_response1["iris_response"][0]
area_sg_load1 = iris_response_load1.area_sg
area_sji_load1 = iris_response_load1.area_sji

# Version 2
sav_file_path2 = Path(ROOTDIR) / "idl_iris_get_response_20130903_new_version002.sav"
test_iris_response2 = scipy.io.readsav(sav_file_path2, python_dict=True, verbose=False)
iris_response_load2 = test_iris_response2["iris_response"][0]
area_sg_load2 = iris_response_load2.area_sg
area_sji_load2 = iris_response_load2.area_sji

# Version 3
sav_file_path3 = Path(ROOTDIR) / "idl_iris_get_response_20130903_new_version003.sav"

test_iris_response3 = scipy.io.readsav(sav_file_path3, python_dict=True, verbose=False)
iris_response_load3 = test_iris_response3["iris_response"][0]
area_sg_load3 = iris_response_load3.area_sg
area_sji_load3 = iris_response_load3.area_sji

# Version 4
sav_file_path4 = Path(ROOTDIR) / "idl_iris_get_response_20130903_new_version004.sav"
test_iris_response4 = scipy.io.readsav(sav_file_path4, python_dict=True, verbose=False)
iris_response_load4 = test_iris_response4["iris_response"][0]
area_sg_load4 = iris_response_load4.area_sg
area_sji_load4 = iris_response_load4.area_sji

# Version 5
sav_file_path5 = Path(ROOTDIR) / "idl_iris_get_response_20130903_new_version005.sav"
test_iris_response5 = scipy.io.readsav(sav_file_path5, python_dict=True, verbose=False)
iris_response_load5 = test_iris_response5["iris_response"][0]
area_sg_load5 = iris_response_load5.area_sg
area_sji_load5 = iris_response_load5.area_sji

# Version 6
sav_file_path6 = Path(ROOTDIR) / "idl_iris_get_response_20130903_new_version006.sav"
test_iris_response6 = scipy.io.readsav(sav_file_path6, python_dict=True, verbose=False)
iris_response_load6 = test_iris_response6["iris_response"][0]
area_sg_load6 = iris_response_load6.area_sg
area_sji_load6 = iris_response_load6.area_sji


def test_get_iris_response_version1(iris_response_v1):
    np_test.assert_almost_equal(iris_response_v1["AREA_SG"].value, area_sg_load1, decimal=6)
    np_test.assert_almost_equal(iris_response_v1["AREA_SJI"].value, area_sji_load1, decimal=6)


def test_get_iris_response_version2(iris_response_v2):
    np_test.assert_almost_equal(iris_response_v2["AREA_SG"].value, area_sg_load2, decimal=6)
    np_test.assert_almost_equal(iris_response_v2["AREA_SJI"].value, area_sji_load2, decimal=6)


@pytest.mark.xfail
def test_get_iris_response_version3(iris_response_v3):
    np_test.assert_almost_equal(iris_response_v3["AREA_SG"].value, area_sg_load3)
    np_test.assert_almost_equal(iris_response_v3["AREA_SJI"].value, area_sji_load3)


@pytest.mark.xfail
def test_get_iris_response_version4(iris_response_v4):
    np_test.assert_almost_equal(iris_response_v4["AREA_SG"].value, area_sg_load4)
    np_test.assert_almost_equal(iris_response_v4["AREA_SJI"].value, area_sji_load4)


@pytest.mark.xfail
def test_get_iris_response_version5(iris_response_v5):
    np_test.assert_almost_equal(iris_response_v5["AREA_SG"].value, area_sg_load5)
    np_test.assert_almost_equal(iris_response_v5["AREA_SJI"].value, area_sji_load5)


@pytest.mark.xfail
def test_get_iris_response_version6(iris_response_v6):
    np_test.assert_almost_equal(iris_response_v6["AREA_SG"].value, area_sg_load6)
    np_test.assert_almost_equal(iris_response_v6["AREA_SJI"].value, area_sji_load6)


@pytest.mark.parametrize(
    ("input_arrays", "expected_array"),
    [
        ([time_obs, time_cal_coeffs0, cal_coeffs0], iris_fit_expected0),
        ([time_obs, time_cal_coeffs0, cal_coeffs1], iris_fit_expected1),
        ([time_obs, time_cal_coeffs0, cal_coeffs2], iris_fit_expected2),
        ([time_obs, time_cal_coeffs1, cal_coeffs3], iris_fit_expected3),
        ([time_obs, time_cal_coeffs1, cal_coeffs4], iris_fit_expected4),
        ([time_obs, time_cal_coeffs1, cal_coeffs5], iris_fit_expected5),
        ([time_obs, time_cal_coeffs1, cal_coeffs6], iris_fit_expected6),
    ],
)
def test_fit_iris_xput(input_arrays, expected_array):
    np_test.assert_almost_equal(
        utils.fit_iris_xput(input_arrays[0], input_arrays[1], input_arrays[2]),
        expected_array,
        decimal=6,
    )


@pytest.mark.xfail
def test_fixed_time():
    sav_file_path6 = Path(ROOTDIR) / "test_response.sav"
    response = scipy.io.readsav(sav_file_path6, python_dict=True, verbose=False)
    response = response["test"][0]
    area_sg = response.area_sg
    area_sji = response.area_sji
    python_response = utils.get_iris_response(parse_time("2013-08-31"), response_version=6)
    np_test.assert_almost_equal(python_response["AREA_SG"].value, area_sg)
    np_test.assert_almost_equal(python_response["AREA_SJI"].value, area_sji)



"""
Configuration file for the Sphinx documentation builder.
"""

import datetime
from pathlib import Path

from sunpy_sphinx_theme import PNG_ICON

from irispy import __version__

# -- Project information -----------------------------------------------------
project = "irispy-lmsal"
author = "IRIS Instrument Team"
copyright = f"{datetime.datetime.now(datetime.timezone.utc).year}, {author}"  # NOQA: A001
release = __version__
is_development = ".dev" in __version__

# -- General configuration ---------------------------------------------------
extensions = [
    "sphinx_copybutton",
    "sphinx_design",
    "sphinx_automodapi.automodapi",
    "sphinx_automodapi.smart_resolver",
    "sphinx_changelog",
    "sphinx_gallery.gen_gallery",
    "sphinx.ext.autodoc",
    "sphinx.ext.coverage",
    "sphinx.ext.doctest",
    "sphinx.ext.inheritance_diagram",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
]
automodapi_toctreedirnm = "generated/api"
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]
source_suffix = ".rst"
master_doc = "index"
default_role = "obj"

# -- Options for sphinx-copybutton ---------------------------------------------
# Python Repl + continuation, Bash, ipython and qtconsole + continuation, jupyter-console + continuation
copybutton_prompt_text = r">>> |\.\.\. |\$ |In \[\d*\]: | {2,5}\.\.\.: | {5,8}: "
copybutton_prompt_is_regexp = True

# -- Options for intersphinx extension ---------------------------------------
intersphinx_mapping = {
    "python": (
        "https://docs.python.org/3/",
        (None, "http://www.astropy.org/astropy-data/intersphinx/python3.inv"),
    ),
    "numpy": (
        "https://numpy.org/doc/stable/",
        (None, "http://www.astropy.org/astropy-data/intersphinx/numpy.inv"),
    ),
    "scipy": (
        "https://docs.scipy.org/doc/scipy/reference/",
        (None, "http://www.astropy.org/astropy-data/intersphinx/scipy.inv"),
    ),
    "matplotlib": ("https://matplotlib.org/stable", None),
    "aiapy": ("https://aiapy.readthedocs.io/en/stable/", None),
    "astropy": ("https://docs.astropy.org/en/stable/", None),
    "astroquery": ("https://astroquery.readthedocs.io/en/latest/", None),
    "drms": ("https://docs.sunpy.org/projects/drms/en/stable/", None),
    "mpl_animators": ("https://docs.sunpy.org/projects/mpl-animators/en/stable/", None),
    "pandas": ("https://pandas.pydata.org/pandas-docs/stable/", None),
    "parfive": ("https://parfive.readthedocs.io/en/stable/", None),
    "reproject": ("https://reproject.readthedocs.io/en/stable/", None),
    "skimage": ("https://scikit-image.org/docs/stable/", None),
    "sqlalchemy": ("https://docs.sqlalchemy.org/en/latest/", None),
    "sunkit_image": ("https://docs.sunpy.org/projects/sunkit-image/en/stable/", None),
    "sunkit_instruments": (
        "https://docs.sunpy.org/projects/sunkit-instruments/en/stable/",
        None,
    ),
    "asdf": ("https://asdf.readthedocs.io/en/stable/", None),
    "sunpy": ("https://docs.sunpy.org/en/stable/", None),
    "sunraster": ("https://docs.sunpy.org/projects/sunraster/en/stable/", None),
    "zeep": ("https://docs.python-zeep.org/en/stable/", None),
}

# -- Options for HTML output -------------------------------------------------
html_theme = "sunpy"
graphviz_output_format = "svg"
sphinx_gallery_conf = {
    "backreferences_dir": Path("generated") / "modules",
    "filename_pattern": "^((?!skip_).)*$",
    "examples_dirs": Path("..") / "examples",
    "within_subsection_order": "ExampleTitleSortKey",
    "gallery_dirs": Path("generated") / "gallery",
    "default_thumb_file": PNG_ICON,
    "abort_on_example_error": False,
    "plot_gallery": "True",
    "remove_config_comments": True,
    "doc_module": ("sunpy"),
    "only_warn_on_example_error": True,
    "matplotlib_animations": True,
}



import numpy as np

import sunpy.map

AXIS = [
    (
        "custom:pos.helioprojective.lon",
        "custom:pos.helioprojective.lat",
        "time",
        "custom:CUSTOM",
        "custom:CUSTOM",
        "custom:CUSTOM",
        "custom:CUSTOM",
        "custom:CUSTOM",
        "custom:CUSTOM",
        "custom:CUSTOM",
        "custom:CUSTOM",
        "custom:CUSTOM",
    ),
    ("custom:pos.helioprojective.lon", "custom:pos.helioprojective.lat"),
    ("custom:pos.helioprojective.lon", "custom:pos.helioprojective.lat"),
]


def test_world_axis_physical_types_sjicube_2832(sjicube_2832):
    assert np.all(sjicube_2832.shape == (10, 40, 37))
    assert sjicube_2832.array_axis_physical_types == AXIS


def test_world_axis_physical_types_sjicube_2796(sjicube_2796):
    assert np.all(sjicube_2796.shape == (62, 40, 37))
    assert sjicube_2796.array_axis_physical_types == AXIS


def test_world_axis_physical_types_sjicube_1400(sjicube_1400):
    assert np.all(sjicube_1400.shape == (62, 40, 37))
    assert sjicube_1400.array_axis_physical_types == AXIS


def test_world_axis_physical_types_sjicube_1330(sjicube_1330):
    assert np.all(sjicube_1330.shape == (52, 40, 37))
    assert sjicube_1330.array_axis_physical_types == AXIS


def test_to_map(sjicube_1330):
    # Basic smoke tests
    output = sjicube_1330.to_maps(0)
    assert isinstance(output, sunpy.map.GenericMap)
    assert output.data.shape == (40, 37)
    assert output.reference_date is not None

    output = sjicube_1330.to_maps([0, 2])
    assert isinstance(output, sunpy.map.mapsequence.MapSequence)
    assert output.data.shape == (40, 37, 2)
    assert np.all([output.reference_date is not None for output in output])

    output = sjicube_1330.to_maps(range(1, 3))
    assert isinstance(output, sunpy.map.mapsequence.MapSequence)
    assert output.data.shape == (40, 37, 2)
    assert np.all([output.reference_date is not None for output in output])

    output = sjicube_1330.to_maps(range(0, 12, 4))
    assert isinstance(output, sunpy.map.mapsequence.MapSequence)
    assert output.data.shape == (40, 37, 3)
    assert np.all([output.reference_date is not None for output in output])

    output = sjicube_1330.to_maps()
    assert isinstance(output, sunpy.map.mapsequence.MapSequence)
    assert output.data.shape == (40, 37, 52)
    assert np.all([output.reference_date is not None for output in output])



import textwrap
import warnings

import matplotlib.pyplot as plt

from sunpy import log as logger
from sunpy.map import Map
from sunpy.util.exceptions import SunpyMetadataWarning
from sunraster import SpectrogramCube

from irispy.utils import calculate_dust_mask
from irispy.visualization import IRISPlotter, set_axis_properties

__all__ = ["AIACube", "SJICube"]


class SJICube(SpectrogramCube):
    """
    Class representing SJI Image described by a single WCS.

    Parameters
    ----------
    data : `numpy.ndarray`
        The array holding the actual data in this object.
    wcs : `ndcube.wcs.wcs.WCS`
        The WCS object containing the axes' information
    unit : `astropy.unit.Unit` or `str`
        Unit for the dataset.
        Strings that can be converted to a Unit are allowed.
    meta : dict-like object
        Additional meta information about the dataset.
    uncertainty : any type, optional
        Uncertainty in the dataset. Should have an attribute uncertainty_type
        that defines what kind of uncertainty is stored, for example "std"
        for standard deviation or "var" for variance. A metaclass defining
        such an interface is NDUncertainty - but isn't mandatory. If the
        uncertainty has no such attribute the uncertainty is stored as
        UnknownUncertainty.
        Defaults to None.
    mask : any type, optional
        Mask for the dataset. Masks should follow the numpy convention
        that valid data points are marked by False and invalid ones with True.
        Defaults to None.
    copy : `bool`, optional
        Indicates whether to save the arguments as copy. True copies every
        attribute before saving it while False tries to save every parameter
        as reference. Note however that it is not always possible to save the
        input as reference.
        Default is False.
    scaled : `bool`, optional
        Indicates if the data has been scaled.
    """

    def __init__(
        self,
        data,
        wcs,
        *,
        uncertainty=None,
        unit=None,
        meta=None,
        mask=None,
        copy=False,
        scaled=None,
        **kwargs,
    ) -> None:
        self.scaled = scaled
        self.dust_masked = False
        self._basic_wcs = kwargs.pop("_basic_wcs") if "_basic_wcs" in kwargs else None
        super().__init__(
            data,
            wcs,
            uncertainty=uncertainty,
            mask=mask,
            meta=meta,
            unit=unit,
            copy=copy,
            **kwargs,
        )

    def __repr__(self) -> str:
        return f"{object.__repr__(self)}\n{self!s}"

    def __str__(self) -> str:
        if self.wcs.world_n_dim == 2:
            instance_start = self.global_coords.get("Time (UTC)")
            instance_end = None
        else:
            instance_start = self.wcs.pixel_to_world(0, 0, 0)[-1]
            instance_end = self.wcs.pixel_to_world(0, 0, self.data.shape[0] - 1)[-1]
        return textwrap.dedent(
            f"""
            SJICube
            -------
            Observatory:           {self.meta.get("TELESCOP", "IRIS")}
            Instrument:            {self.meta.get("INSTRUME")}
            Bandpass:              {self.meta.get("TWAVE1")}
            Obs. Start:            {self.meta.get("STARTOBS")}
            Obs. End:              {self.meta.get("ENDOBS")}
            Instance Start:        {instance_start}
            Instance End:          {instance_end}
            Total Frames in Obs.:  {self.meta.get("NBFRAMES")}
            IRIS Obs. id:          {self.meta.get("OBSID")}
            IRIS Obs. Description: {self.meta.get("OBS_DESC")}
            Axis Types:            {self.array_axis_physical_types}
            Roll:                  {self.meta.get("SAT_ROT")}
            Cube dimensions:       {self.shape}
            """,
        )

    def __getitem__(self, item):
        sliced_self = super().__getitem__(item)
        sliced_self.scaled = self.scaled
        if self._basic_wcs is not None:
            sliced_self._basic_wcs = self._basic_wcs[item]
        return sliced_self

    def plot(self, *args, **kwargs):
        cmap = kwargs.get("cmap")
        if not cmap:
            try:
                cmap = plt.get_cmap(name=f"irissji{int(self.meta['TWAVE1'])}")
            except Exception as e:  # NOQA: BLE001
                logger.debug(e)
                cmap = "viridis"
        kwargs["cmap"] = cmap
        ax = IRISPlotter(ndcube=self).plot(*args, **kwargs)
        set_axis_properties(ax)
        return ax

    def apply_dust_mask(self, *, undo=False):
        """
        Applies or undoes an update of the mask with the dust particles
        positions.

        Rewrite self.mask with/without the dust positions.

        Parameters
        ----------
        undo: `bool`
            If False, dust particles positions mask will be applied.
            If True, dust particles positions mask will be removed.
            Default=False
        """
        dust_mask = calculate_dust_mask(self.data)
        if undo:
            # If undo kwarg IS set, unmask dust pixels.
            self.mask[dust_mask] = False
            self.dust_masked = False
        else:
            # If undo kwarg is NOT set, mask dust pixels.
            self.mask[dust_mask] = True
            self.dust_masked = True

    @property
    def basic_wcs(self):
        """
        Returns a standard WCS instead of gWCS.
        """
        return self._basic_wcs

    def to_maps(self, index: int | list[int] | None = None):
        """
        Return SunPy Maps for the requested frame(s).

        Parameters
        ----------
        index : int, list, optional
            The index of the SJI steps you want.
            By default None which will return the entire cube as a map sequence.

        Returns
        -------
        `sunpy.map.Map` or `sunpy.map.MapSequence`
            A single Map if index is an int, otherwise a MapSequence.
        """
        if isinstance(index, int):
            idx_list = [index]
        elif index is None:
            idx_list = range(self.data.shape[0])
        else:
            idx_list = index

        # We can shortcut if the Cube has been reduced to a 2D slice
        if self.wcs.world_n_dim == 2:
            # TODO: Missing metadata
            return Map(self.data, self.wcs)
        data_wcs = ((self.data[i], self.basic_wcs[i]) for i in idx_list)
        times_iso = (self.wcs.pixel_to_world(0, 0, i)[-1].utc.isot for i in idx_list)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", SunpyMetadataWarning)
            maps = Map(data_wcs, sequence=True)
        for m, t in zip(maps, times_iso, strict=True):
            m.meta["DATE-OBS"] = t
            m.meta["INSTRUME"] = self.meta.get("INSTRUME", "SJI")
            m.meta["TELESCOP"] = self.meta.get("TELESCOP", "IRIS")
            m.meta["EXPTIME"] = self.meta.get("EXPTIME", 0.0)
            m.meta["TWAVE1"] = self.meta.get("TWAVE1")
            m.plot_settings["cmap"] = f"irissji{int(self.meta['TWAVE1'])}"
        return maps[0] if isinstance(index, int) else maps


class AIACube(SJICube):
    """
    Subclass of the SJICube.

    It is the same outside of the name.
    """

    def __str__(self) -> str:
        return super().__str__().replace("SJICube", "AIACube")



import numpy as np
import numpy.testing as np_test
import pytest

from astropy import units as u

from sunpy.time import parse_time

from irispy import utils

data_dust = np.array(
    [
        [[-1, 2, -3, 4], [2, -200, 5, 3], [0, 1, 2, -300]],
        [[2, -200, 5, 1], [10, -5, 2, 2], [10, -3, 3, 0]],
    ],
)
dust_mask_expected = np.array(
    [
        [
            [True, True, True, True],
            [True, True, True, True],
            [True, True, False, False],
        ],
        [[True, True, True, False], [True, True, True, True], [True, True, True, True]],
    ],
)


@pytest.mark.parametrize(
    ("test_input", "expected_output"),
    [
        ({"detector type": "FUV1"}, "FUV"),
        ({"detector type": "NUV"}, "NUV"),
        ({"detector type": "SJI"}, "SJI"),
    ],
)
def test_get_detector_type(test_input, expected_output):
    assert utils.get_detector_type(test_input) == expected_output


@pytest.mark.parametrize(("input_array", "expected_array"), [(data_dust, dust_mask_expected)])
def test_calculate_dust_mask(input_array, expected_array):
    np_test.assert_array_equal(utils.calculate_dust_mask(input_array), expected_array)


def test_get_iris_response_version():
    # For now just checks the code runs
    start_obs = parse_time("2013-07-20T17:10:23")
    obs_wavelength = np.linspace(1400.5, 1404.9915000926703, num=692, endpoint=True)
    utils.get_interpolated_effective_area(
        start_obs,
        response_version=6,
        detector_type="FUV",
        obs_wavelength=obs_wavelength * u.Angstrom,
    )



"""
This module provides constants used elsewhere.
"""

import astropy.units as u

__all__ = [
    "BAD_PIXEL_VALUE_SCALED",
    "BAD_PIXEL_VALUE_UNSCALED",
    "DN_UNIT",
    "RADIANCE_UNIT",
    "READOUT_NOISE",
    "SLIT_WIDTH",
]

# The following value is only appropriate for byte scaled images
BAD_PIXEL_VALUE_SCALED = -200
# The following value is only appropriate for unscaled images
BAD_PIXEL_VALUE_UNSCALED = -32768
# Define some properties of IRIS detectors. Source: IRIS instrument paper.
DETECTOR_GAIN = {"NUV": 18.0, "FUV": 6.0, "SJI": 18.0}
DETECTOR_YIELD = {"NUV": 1.0, "FUV": 1.5, "SJI": 1.0}
DN_UNIT = {
    "NUV": u.def_unit("DN_IRIS_NUV", DETECTOR_GAIN["NUV"] / DETECTOR_YIELD["NUV"] * u.photon),
    "FUV": u.def_unit("DN_IRIS_FUV", DETECTOR_GAIN["FUV"] / DETECTOR_YIELD["FUV"] * u.photon),
    "SJI": u.def_unit("DN_IRIS_SJI", DETECTOR_GAIN["SJI"] / DETECTOR_YIELD["SJI"] * u.photon),
    "SJI_UNSCALED": u.def_unit("DN_IRIS_SJI_UNSCALED", u.ct),
}

READOUT_NOISE = {
    "NUV": 1.2 * DN_UNIT["NUV"],
    "FUV": 3.1 * DN_UNIT["FUV"],
    "SJI": 1.2 * DN_UNIT["SJI"],
}
RADIANCE_UNIT = u.erg / u.cm**2 / u.s / u.steradian / u.Angstrom
SLIT_WIDTH = 0.33 * u.arcsec
SPECTRAL_BAND = {
    "1343": "FUV",
    "2786": "NUV",
    "2814": "NUV",
    "2826": "NUV",
    "2830": "NUV",
    "2832": "NUV",
    "C II 1336": "FUV",
    "Cl I 1352": "FUV",
    "Fe XII 1349": "FUV",
    "Mg II k 2796": "NUV",
    "O I 1356": "FUV",
    "Si IV 1394": "FUV",
    "Si IV 1403": "FUV",
}



# NOTE: First try _dev.scm_version if it exists and setuptools_scm is installed
# This file is not included in irispy-lmsal wheels/tarballs, so otherwise it will
# fall back on the generated _version module.
try:
    try:
        from ._dev.scm_version import version
    except ImportError:
        from ._version import version
except Exception:  # NOQA: BLE001
    import warnings

    warnings.warn(
        f"could not determine {__name__.split('.')[0]} package version; this indicates a broken installation",
        stacklevel=3,
    )
    del warnings
    version = "0.0.0"

from packaging.version import parse as _parse

_version = _parse(version)
major, minor, bugfix = [*_version.release, 0][:3]
release = not _version.is_devrelease



from mpl_animators import ArrayAnimatorWCS

import astropy.units as u

import sunpy.visualization.colormaps as cm  # NOQA: F401
from ndcube.visualization.mpl_plotter import MatplotlibPlotter
from ndcube.visualization.mpl_sequence_plotter import MatplotlibSequencePlotter, SequenceAnimator

__all__ = ["IRISArrayAnimatorWCS", "IRISPlotter", "IRISSequencePlotter"]


LAT_LABELS = [
    "custom:pos.helioprojective.lat",
    "hplt-tan",
    "hplt",
    "lat",
    "latitude",
]
LON_LABELS = [
    "custom:pos.helioprojective.lon",
    "hpln-tan",
    "hpln",
    "lon",
    "longitude",
]
WAVELENGTH_LABELS = ["wavelength", "wave", "em.wl"]


def set_axis_properties(ax):
    """
    Set the axis colors and labels for IRIS SJI and Raster data.
    """
    if isinstance(ax, ArrayAnimatorWCS):
        ax = ax.axes
    for axis in ax.coords:
        if axis.default_label.lower() in WAVELENGTH_LABELS:
            axis.set_format_unit(u.nm)
            axis.set_major_formatter("x.x")
            axis.set_axislabel("Wavelength [$\\mathrm{nm}$]")
        elif axis.default_label.lower() in LAT_LABELS:
            _set_axis_properties(axis, "Helioprojective Latitude [arcsec]", "red")
        elif axis.default_label.lower() in LON_LABELS:
            _set_axis_properties(axis, "Helioprojective Longitude [arcsec]", "black")


def _set_axis_properties(axis, label, color):
    """
    Set the axis colors and labels for IRIS SJI and Raster data.

    Parameters
    ----------
    axis : `~astropy.visualization.wcsaxes.core.WCSAxes`
        The axis to set the colors and labels for.
    label : str
        The label to use for the axis.
    color : str
        The color to use for the axis label.
    """
    axis.set_ticklabel(color, fontsize=8)
    axis.set_axislabel(label, color=color, fontsize=8)


class Plot2DMixin:
    def update_plot_2d(self, val, im, slider):
        super().update_plot_2d(val, im, slider)
        set_axis_properties(self.axes)


class IRISArrayAnimatorWCS(Plot2DMixin, ArrayAnimatorWCS):
    pass


class IRISSequenceAnimator(Plot2DMixin, SequenceAnimator):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault("slider_labels", ["Raster Step", "Scan Number"])
        super().__init__(*args, **kwargs)


class IRISPlotter(MatplotlibPlotter):
    def _animate_cube(
        self,
        wcs,
        plot_axes=None,
        axes_coordinates=None,
        axes_units=None,
        data_unit=None,
        **kwargs,
    ):
        # This is a copy of the super method, but with the replacement
        # of the ArrayAnimatorWCS with IRISArrayAnimatorWCS.
        data, wcs, plot_axes, coord_params = self._prep_animate_args(wcs, plot_axes, axes_units, data_unit)
        ax = IRISArrayAnimatorWCS(data, wcs, plot_axes, coord_params=coord_params, **kwargs)
        # We need to modify the visible axes after the axes object has been created.
        # This call affects only the initial draw
        self._apply_axes_coordinates(ax.axes, axes_coordinates)
        # This changes the parameters for future iterations
        for hidden in self._not_visible_coords(ax.axes, axes_coordinates):
            param = ax.coord_params.get(hidden, {})
            param["ticks"] = False
            ax.coord_params[hidden] = param
        return ax


class IRISSequencePlotter(MatplotlibSequencePlotter):
    def animate(self, sequence_axis_coords=None, sequence_axis_unit=None, **kwargs):
        """
        Animate the `~ndcube.NDCubeSequence` with the sequence axis as a
        slider.

        Keyword arguments are passed to
        `ndcube.visualization.mpl_plotter.MatplotlibPlotter.plot` and therefore only
        apply to cube axes, not the sequence axis.
        See that method's docstring for definition of keyword arguments.

        Parameters
        ----------
        sequence_axis_coords: `str` optional
            The name of the coordinate in `~ndcube.NDCubeSequence.sequence_axis_coords`
            to be used as the slider pixel values.
            If None, array indices will be used.

        sequence_axis_unit: `astropy.units.Unit` or `str`, optional
            The unit in which the sequence_axis_coordinates should be displayed.
            If None, the default unit will be used.
        """
        return IRISSequenceAnimator(
            self._ndcube, sequence_axis_coords=sequence_axis_coords, sequence_axis_unit=sequence_axis_unit, **kwargs
        )



"""
===========
A quick SJI
===========

In this example we will show how to plot a South Pole SJI dataset.
"""

import matplotlib.pyplot as plt
import pooch

from irispy.io import read_files

###############################################################################
# We start with getting the data.
# This is done by downloading the data from the IRIS archive.
#
# In this case, we will use ``pooch`` so to keep this example self-contained
# but using your browser will also work.

sji_filename = pooch.retrieve(
    "https://www.lmsal.com/solarsoft/irisa/data/level2_compressed/2023/02/11/20230211_083601_3880012095/iris_l2_20230211_083601_3880012095_SJI_2832_t000.fits.gz",
    known_hash="282595a2c0a5400ef98d1df0172618483dc74b5b56ff16a0f388723363c6f0f1",
)

###############################################################################
# We will now open the slit-jaw imager (SJI) file we just downloaded.

sji_2832 = read_files(sji_filename, memmap=False)
# Printing will give us an overview of the SJI dataset.
print(sji_2832)

###############################################################################
# We will now plot the IRIS SJI data.
#
# You can also change the axis labels and ticks if you so desire.
# `WCSAxes provides us an API we can use. <https://docs.astropy.org/en/stable/visualization/wcsaxes/index.html>`__

# Note that the .get_animation() is used to animate this example and is not required normally.
ax = sji_2832.plot().get_animation()
plt.title(f"IRIS SJI {sji_2832.meta['TWAVE1']}", pad=25)

###############################################################################
# Finally we will output a frame of the SJI into a sunpy Map.

sji_map = sji_2832.to_maps(0)
print(sji_map)

###############################################################################
# We can now plot the SJI Map using sunpy Map's plotting capabilities.

fig = plt.figure()
ax = fig.add_subplot(projection=sji_map.wcs)
sji_map.plot(axes=ax)

plt.show()



"""
============
irispy-lmsal
============

``irispy-lmsal`` is a library that provides the tools to read in and analyze data from Interface Region Imaging Spectrograph (IRIS).
"""

from .sji import *
from .spectrograph import *
from .version import version as __version__  # NOQA: F401



"""
=======================
Umbral flashes analysis
=======================

In this example, we are going to work with IRIS data to
study an example of a dynamical phenomena called `umbral flashes <https://ui.adsabs.harvard.edu/abs/1973SoPh...30..403M>`__.
"""

import matplotlib.dates as mdates
import matplotlib.pyplot as plt
import pooch

import astropy.units as u
from astropy.coordinates import SpectralCoord
from astropy.visualization import time_support

from irispy.io import read_files

time_support()

###############################################################################
# We will start by getting some data from the IRIS archive.
#
# In this case, we will use ``pooch`` so to keep this example self-contained
# but using your browser will also work.
#
# Using the url: http://www.lmsal.com/solarsoft/irisa/data/level2_compressed/2013/09/02/20130902_163935_4000255147/
# we are after the 1400 Slit-Jaw and the raster sequence (~900 MB).

raster_filename = pooch.retrieve(
    "http://www.lmsal.com/solarsoft/irisa/data/level2_compressed/2013/09/02/20130902_163935_4000255147/iris_l2_20130902_163935_4000255147_raster.tar.gz",
    known_hash="e0fdbe63b6ff88856d97dff1fd4c3f62bf0bb4a85c025930c36fd69086ed0533",
)
sji_filename = pooch.retrieve(
    "http://www.lmsal.com/solarsoft/irisa/data/level2_compressed/2013/09/02/20130902_163935_4000255147/iris_l2_20130902_163935_4000255147_SJI_1400_t000.fits.gz",
    known_hash="e04eedbe7fd7e5e173b6a9308db3e45991634a4f8d9b245a334d5c492d85d0ed",
)

###############################################################################
# Now to open the files using ``irispy-lmsal``.

# Note that when ``memmap=True``, the data values are read from the FITS file
# directly without the scaling to Float32 (via "b_zero" and "b_scale"),
# the data values are no longer in DN, but in scaled integer units that start at -2$^{16}$/2.

raster = read_files(raster_filename, memmap=True, uncertainty=False)
sji_1400 = read_files(sji_filename, memmap=True, uncertainty=False)

###############################################################################
# We are after the Mg II k and C II lines, which we can select using keys.
# Then we will produce a space-time image of the Mg II k3 line.

mg_ii = raster["Mg II k 2796"][0]
c_ii = raster["C II 1336"][0]

# Instead of using a pixel index, we can crop the data in wavelength space.
lower_corner = [SpectralCoord(279.63, unit=u.nm), None]
upper_corner = [SpectralCoord(279.63, unit=u.nm), None]
mg_crop = mg_ii.crop(lower_corner, upper_corner)

fig = plt.figure()
ax = fig.add_subplot(111, projection=mg_crop.wcs)
mg_crop.plot(axes=ax)

###############################################################################
# The middle section between 60"-75" is on the umbra of a sunspot, even though
# it is not obvious from this image. One can see very clearly the umbral oscillations,
# with a clear regular pattern of dark/bright streaks.
#
# Let us now load the 1400 SJI for context.

plt.figure()
sji_1400[0].plot(vmin=-32000, vmax=-30000)
plt.title("1400 SJI")

###############################################################################
# The slit pixel, "220" is a location on the sunspot's umbra.
# Let us plot the k3 intensity (spectral pixel 103 of ``mg_ii``) and the
# core of the brightest C II line (spectral pixel 90 of ``c_ii``) vs
# time in minutes (showing first ~10 minutes only)

mg_ii_times = mg_ii.time[:200]
c_ii_times = c_ii.time[:200]

plt.figure()
plt.plot(mg_ii_times, mg_ii.data[:200, 220, 103], label="Mg II k3")
(ax,) = plt.plot(c_ii_times, c_ii.data[:200, 220, 90], label="C II")
plt.legend()
plt.ylabel("DN (Memory Mapped Value)")
plt.xlabel("Time (UTC)")
ax.axes.xaxis.set_major_formatter(mdates.ConciseDateFormatter(ax.axes.xaxis.get_major_locator()))
# Rotates and right-aligns the x labels so they don't crowd each other.
for label in ax.axes.get_xticklabels(which="major"):
    label.set(rotation=30, horizontalalignment="right")

plt.tight_layout()

###############################################################################
# Imagine now you wanted to compare these oscillations with
# the intensity from the SJI. The SJI images are typically
# taken at a different cadence, so you need get the corresponding
# times for the 1400 SJI.

# We will take the first 50 to cut down on the size of the data for this example.
times_sji = sji_1400.time[:50]

###############################################################################
# Now we can plot both spectral lines and SJI for a pixel close to the slit
# at the same Y position (pre-worked out to be at index 220).

plt.figure()
plt.plot(mg_ii_times, mg_ii.data[:200, 220, 103], label="Mg II k3")
plt.plot(c_ii_times, c_ii.data[:200, 220, 90], label="C II")
(ax,) = plt.plot(times_sji, sji_1400.data[:50, 190, 220], label="1400 SJI")
plt.legend()
plt.ylabel("Counts (Memory Mapped Value)")
plt.xlabel("Time (UTC)")
ax.axes.xaxis.set_major_formatter(mdates.ConciseDateFormatter(ax.axes.xaxis.get_major_locator()))
# Rotates and right-aligns the x labels so they don't crowd each other.
for label in ax.axes.get_xticklabels(which="major"):
    label.set(rotation=30, horizontalalignment="right")

plt.tight_layout()

plt.show()

###############################################################################
# You are now ready to explore all the correlations, anti-correlations,
# and phase differences.



"""
=============================
Dealing with IRIS v34 rasters
=============================

In this example we will show how ``irispy-lmsal`` deals with a v34 dataset by default and how to
undo that if you so desire.

These v34 are scans which raster from west to east instead of the default east to west.
"""

import matplotlib.pyplot as plt
import pooch

import astropy.units as u
from astropy.coordinates import SkyCoord, SpectralCoord
from astropy.wcs.utils import wcs_to_celestial_frame

from sunpy.coordinates.frames import Helioprojective

from irispy.io import read_files

###############################################################################
# We start by downloading a raster v34 dataset from the IRIS archive.
#
# In this case, we will use ``pooch`` to keep this example self-contained
# but using your browser will also work.
#
# This is a very large sparse 64-step raster with a FOV of 63"x175" for
# a total of 8 complete raster scans.

raster_filename = pooch.retrieve(
    "https://www.lmsal.com/solarsoft/irisa/data/level2_compressed/2025/03/28/20250328_225628_3400109360/iris_l2_20250328_225628_3400109360_raster.tar.gz",
    known_hash="ce46e93a8962b04da344c203c4f75a8d7411d1f587e25d728a6ca8398208de1a",
)

###############################################################################
# We will now open the raster file we just downloaded.
# By default, irispy-lmsal will read the v34 data, flipping the data so that it
# is in the same orientation as normal IRIS data and adjust the WCS accordingly.

raster = read_files(raster_filename, memmap=False)
# We will also undo the v34 handling and read the data as is.
raster_unflipped = read_files(raster_filename, memmap=False, revert_v34=True)

# Printing will give us an overview of the file.
print(raster)

###############################################################################
# We will now pull the 'Mg II k 2796' spectral line data from the raster.

mg_ii_k = raster["Mg II k 2796"]
mg_ii_k_unflipped = raster_unflipped["Mg II k 2796"]
print(mg_ii_k)

###############################################################################
# To see the effect of the v34 handling, we will plot a spectroheliogram
# for the Mg II k core wavelength.
#
# We can use the ``crop`` method to get this information, this will
# require a `~.SpectralCoord` object from `astropy.coordinates`.

# None, means that the axis is not cropped
# Since we want one physical coordinate, we will just use the
# same spectral coordinate for both axes.
lower_corner = [SpectralCoord(280, unit=u.nm), None]

mg_spec_crop = mg_ii_k[0].crop(lower_corner, lower_corner)
mg_spec_unflipped_crop = mg_ii_k_unflipped[0].crop(lower_corner, lower_corner)

fig = plt.figure(figsize=(6, 12))
ax = fig.add_subplot(121, projection=mg_spec_crop.wcs)
ax.set_title("v34 flipped")
mg_spec_crop.plot(axes=ax, plot_axes=["x", "y"])

ax2 = fig.add_subplot(122, projection=mg_spec_unflipped_crop.wcs)
ax2.set_title("v34 unflipped")
mg_spec_unflipped_crop.plot(axes=ax2, plot_axes=["x", "y"])
fig.tight_layout()

###############################################################################
# As you can see, the v34 data is flipped in the y-axis and the WCS is
# adjusted accordingly.
#
# The same is true for the times in the raster:

print(f"Flipped time: {mg_ii_k.time[:5]}")
print("*" * 50)
print(f"Unflipped time: {mg_ii_k_unflipped.time[:5]}")

###############################################################################
# Finally we will just see that the spectral profiles are unaffected in either case.

iris_observer = wcs_to_celestial_frame(mg_ii_k[0].wcs.celestial).observer
iris_frame = Helioprojective(observer=iris_observer)
lower_corner = [None, SkyCoord(-909 * u.arcsec, 294 * u.arcsec, frame=iris_frame)]
mg_ii_k_unflipped_spectra = mg_ii_k_unflipped[0].crop(lower_corner, lower_corner)
mg_ii_k_spectra = mg_ii_k[0].crop(lower_corner, lower_corner)

fig = plt.figure()
ax = fig.add_subplot(111, projection=mg_ii_k_unflipped_spectra.wcs)
mg_ii_k_unflipped_spectra.plot(axes=ax, color="red", label="v34 unflipped")
mg_ii_k_spectra.plot(axes=ax, color="black", label="v34 default", linestyle="--")
plt.legend()

###############################################################################

plt.show()



import numpy as np
import numpy.testing as np_test
import pytest

import astropy.units as u

from irispy.utils.constants import DN_UNIT
from irispy.utils.spectrograph import convert_between_dn_and_photons

SOURCE_DATA_DN = np.array([[0.563, 1.132, -1.343], [-0.719, 1.441, 1.566]])
SOURCE_DATA_DN_1 = np.array([[1, 2, 3], [4, 5, 6]])
SOURCE_DATA_PHOTONS_NUV = np.array([[10.134, 20.376, -24.174], [-12.942, 25.938, 28.188]])
SOURCE_DATA_PHOTONS_FUV = np.array([[2.252, 4.528, -5.372], [-2.876, 5.764, 6.264]])
SOURCE_DATA_PHOTONS_SJI_1 = np.array([[18, 36, 54], [72, 90, 108]])


@pytest.mark.parametrize(
    ("data_arrays", "old_unit", "new_unit", "expected_data_arrays", "expected_unit"),
    [
        (
            [SOURCE_DATA_DN, SOURCE_DATA_DN],
            DN_UNIT["FUV"],
            u.photon,
            [SOURCE_DATA_PHOTONS_FUV, SOURCE_DATA_PHOTONS_FUV],
            u.photon,
        ),
        (
            [SOURCE_DATA_DN, SOURCE_DATA_DN],
            DN_UNIT["NUV"],
            u.photon,
            [SOURCE_DATA_PHOTONS_NUV, SOURCE_DATA_PHOTONS_NUV],
            u.photon,
        ),
        (
            [SOURCE_DATA_DN_1, SOURCE_DATA_DN_1],
            DN_UNIT["SJI"],
            u.photon,
            [SOURCE_DATA_PHOTONS_SJI_1, SOURCE_DATA_PHOTONS_SJI_1],
            u.photon,
        ),
        (
            [SOURCE_DATA_PHOTONS_FUV, SOURCE_DATA_PHOTONS_FUV],
            u.photon,
            DN_UNIT["FUV"],
            [SOURCE_DATA_DN, SOURCE_DATA_DN],
            DN_UNIT["FUV"],
        ),
        (
            [SOURCE_DATA_PHOTONS_NUV, SOURCE_DATA_PHOTONS_NUV],
            u.photon,
            DN_UNIT["NUV"],
            [SOURCE_DATA_DN, SOURCE_DATA_DN],
            DN_UNIT["NUV"],
        ),
        (
            [SOURCE_DATA_PHOTONS_SJI_1, SOURCE_DATA_PHOTONS_SJI_1],
            u.photon,
            DN_UNIT["SJI"],
            [SOURCE_DATA_DN_1, SOURCE_DATA_DN_1],
            DN_UNIT["SJI"],
        ),
    ],
)
def test_convert_between_dn_and_photons(data_arrays, old_unit, new_unit, expected_data_arrays, expected_unit):
    output_arrays, output_unit = convert_between_dn_and_photons(data_arrays, old_unit, new_unit)
    for i, output_array in enumerate(output_arrays):
        np_test.assert_allclose(output_array, expected_data_arrays[i])
    assert output_unit == expected_unit



"""
IRISPy test data files.
"""

from pathlib import Path

from astropy.utils.data import get_pkg_data_filename

import irispy

__all__ = ["ROOTDIR", "get_test_filepath"]

ROOTDIR = Path(irispy.__file__).parent / "data" / "test"


def get_test_filepath(filename, **kwargs):
    """
    Return the full path to a test file in the ``data/test`` directory.

    Parameters
    ----------
    filename : `str`
        The name of the file inside the ``data/test`` directory.

    Return
    ------
    filepath : `str`
        The full path to the file.

    See Also
    --------
    astropy.utils.data.get_pkg_data_filename : Get package data filename

    Notes
    -----
    This is a wrapper around `astropy.utils.data.get_pkg_data_filename` which
    sets the ``package`` kwarg to be 'irispy.data.test`.
    """
    return get_pkg_data_filename(filename, package="irispy.data.test", **kwargs)



import numpy as np

import astropy.modeling.models as m
import astropy.units as u
import gwcs
import gwcs.coordinate_frames as cf
from astropy.io import fits
from astropy.time import Time
from astropy.wcs import WCS

import dkist
from dkist.wcs.models import CoupledCompoundModel, VaryingCelestialTransform
from sunpy.coordinates.ephemeris import get_body_heliographic_stonyhurst
from sunpy.coordinates.frames import Helioprojective
from sunpy.map.header_helper import make_fitswcs_header

from irispy.sji import AIACube, SJICube
from irispy.utils import calculate_uncertainty
from irispy.utils.constants import BAD_PIXEL_VALUE_SCALED, BAD_PIXEL_VALUE_UNSCALED, DN_UNIT, READOUT_NOISE

__all__ = ["read_sji_lvl2"]


def _create_gwcs(hdulist: fits.HDUList) -> gwcs.WCS:
    """
    Creates the GWCS object for the SJI file.

    Parameters
    ----------
    hdulist : `astropy.io.fits.HDUList`
        The HDU list of the SJI file.

    Returns
    -------
    `gwcs.WCS`
        GWCS object for the SJI file.
    """
    pc_table = hdulist[1].data[:, hdulist[1].header["PC1_1IX"] : hdulist[1].header["PC2_2IX"] + 1].reshape(-1, 2, 2)
    crval_table = hdulist[1].data[:, hdulist[1].header["XCENIX"] : hdulist[1].header["YCENIX"] + 1]
    crpix_table = [hdulist[0].header["CRPIX1"], hdulist[0].header["CRPIX2"]]
    cdelt = [hdulist[0].header["CDELT1"], hdulist[0].header["CDELT2"]]
    older_dkist = dkist.__version__ < "1.12.0"
    kwargs = {"crpix": crpix_table * u.pixel} if older_dkist else {"crpix_table": crpix_table * u.pixel}
    celestial = VaryingCelestialTransform(
        cdelt=cdelt * u.arcsec / u.pixel,
        pc_table=pc_table * u.pixel,
        crval_table=crval_table * u.arcsec,
        **kwargs,
    )
    base_time = Time(hdulist[0].header["STARTOBS"], format="isot", scale="utc")
    times = hdulist[1].data[:, hdulist[1].header["TIME"]] * u.s
    # We need to account for a non-zero time delta.
    base_time += times[0]
    times -= times[0]
    temporal = m.Tabular1D(
        np.arange(hdulist[1].data.shape[0]) * u.pix,
        lookup_table=times,
        fill_value=np.nan,
        bounds_error=False,
        method="linear",
    )
    forward_transform = CoupledCompoundModel("&", left=celestial, right=temporal)
    celestial_frame = cf.CelestialFrame(
        axes_order=(0, 1),
        unit=(u.arcsec, u.arcsec),
        reference_frame=Helioprojective(observer="earth", obstime=base_time),
        axis_physical_types=[
            "custom:pos.helioprojective.lon",
            "custom:pos.helioprojective.lat",
        ],
        axes_names=("Longitude", "Latitude"),
    )
    temporal_frame = cf.TemporalFrame(Time(base_time), unit=(u.s,), axes_order=(2,), axes_names=("Time (UTC)",))
    output_frame = cf.CompositeFrame([celestial_frame, temporal_frame])
    input_frame = cf.CoordinateFrame(
        axes_order=(0, 1, 2),
        naxes=3,
        axes_type=["PIXEL", "PIXEL", "PIXEL"],
        unit=(u.pix, u.pix, u.pix),
    )
    return gwcs.WCS(forward_transform, input_frame=input_frame, output_frame=output_frame)


def _create_wcs(hdulist):
    """
    This is required as occasionally we need a normal WCS instead of a gWCS due
    to compatibility issues.

    This has been set to have an Earth Observer at the time of the
    observation.
    """
    wcses = []
    base_time = Time(hdulist[0].header["STARTOBS"], format="isot", scale="utc")
    times = hdulist[1].data[:, hdulist[1].header["TIME"]] * u.s
    # We need to account for a non-zero time delta.
    base_time += times[0]
    times -= times[0]
    xcenix_idx = hdulist[1].header["XCENIX"]
    ycenix_idx = hdulist[1].header["YCENIX"]
    pc1_1ix = hdulist[1].header["PC1_1IX"]
    pc1_2ix = hdulist[1].header["PC1_2IX"]
    pc2_1ix = hdulist[1].header["PC2_1IX"]
    pc2_2ix = hdulist[1].header["PC2_2IX"]
    for i in range(hdulist[0].header["NAXIS3"]):
        location = get_body_heliographic_stonyhurst("Earth", (base_time + times[i]).isot)
        observer = Helioprojective(
            hdulist[1].data[i, xcenix_idx] * u.arcsec,
            hdulist[1].data[i, ycenix_idx] * u.arcsec,
            observer=location,
            obstime=base_time + times[i],
        )
        rotation_matrix = np.asanyarray(
            [
                [hdulist[1].data[i, pc1_1ix], hdulist[1].data[i, pc1_2ix]],
                [hdulist[1].data[i, pc2_1ix], hdulist[1].data[i, pc2_2ix]],
            ]
        )
        new_header = make_fitswcs_header(
            data=hdulist[0].data[i].shape,
            coordinate=observer,
            scale=[hdulist[0].header["CDELT1"], hdulist[0].header["CDELT2"]] * u.arcsec / u.pixel,
            rotation_matrix=rotation_matrix,
            instrument="SJI",
            telescope="IRIS",
            observatory="IRIS",
            wavelength=int(hdulist[0].header["TWAVE1"]) * u.AA,
            exposure=hdulist[1].data[i, hdulist[1].header["EXPTIMES"]] * u.second,
            unit=u.DN,
        )
        wcses.append(WCS(new_header))
    return wcses


def read_sji_lvl2(filename, *, uncertainty=False, memmap=False):
    """
    Reads a SINGLE level 2 SJI FITS or the IRIS aligned AIA Cube.

    Does not handle multiple files nor tar files.

    Parameters
    ----------
    filename: `str`
        Filename to read.
    uncertainty : `bool`, optional
        If `True` (not the default), will compute the uncertainty for the data (slower and
        uses more memory). If ``memmap=True``, the uncertainty is never computed.
    memmap : `bool`, optional
        If `True` (not the default), will not load arrays into memory, and will only read from
        the file into memory when needed. This option is faster and uses a
        lot less memory. However, because FITS scaling is not done on-the-fly,
        the data units will be unscaled, not the usual data numbers (DN).

    Returns
    -------
    `irispy.sji.SJICube`
        The data cube, using a gWCS.
    """
    with fits.open(filename, memmap=memmap, do_not_scale_image_data=memmap) as hdulist:
        hdulist.verify("silentfix")
        instrume = hdulist[0].header["INSTRUME"]
        extra_coords = [
            (
                "exposure time",
                0,
                hdulist[1].data[:, hdulist[1].header["EXPTIMES"]] * u.s,
            ),
            (
                "obs_vrix",
                0,
                hdulist[1].data[:, hdulist[1].header["OBS_VRIX"]] * u.m / u.s,
            ),
            (
                "ophaseix",
                0,
                hdulist[1].data[:, hdulist[1].header["OPHASEIX"]] * u.arcsec,
            ),
            ("pztx", 0, hdulist[1].data[:, hdulist[1].header["PZTX"]] * u.arcsec),
            ("pzty", 0, hdulist[1].data[:, hdulist[1].header["PZTY"]] * u.arcsec),
            (
                "slit x position",
                0,
                hdulist[1].data[:, hdulist[1].header["SLTPX1IX"]] * u.arcsec,
            ),
            (
                "slit y position",
                0,
                hdulist[1].data[:, hdulist[1].header["SLTPX2IX"]] * u.arcsec,
            ),
            ("xcenix", 0, hdulist[1].data[:, hdulist[1].header["XCENIX"]] * u.arcsec),
            ("ycenix", 0, hdulist[1].data[:, hdulist[1].header["YCENIX"]] * u.arcsec),
        ]
        data = hdulist[0].data
        data_nan_masked = hdulist[0].data
        out_uncertainty = None
        if memmap:
            data_nan_masked[data == BAD_PIXEL_VALUE_UNSCALED] = 0
            mask = None
            scaled = False
            unit = DN_UNIT["SJI_UNSCALED"]
        else:
            # This is a workaround for the AIA cubes being in int and not float
            mask_value = -200 if np.issubdtype(data.dtype, np.integer) else np.nan
            data_nan_masked[data == BAD_PIXEL_VALUE_SCALED] = mask_value
            mask = data_nan_masked == BAD_PIXEL_VALUE_SCALED
            scaled = True
            unit = DN_UNIT["SJI"]
            if uncertainty and instrume in ["IRIS", "SJI"]:
                out_uncertainty = calculate_uncertainty(data, READOUT_NOISE["SJI"], DN_UNIT["SJI"])
        cube_class = SJICube if instrume in ["IRIS", "SJI"] else AIACube
        map_cube = cube_class(
            data_nan_masked,
            _create_gwcs(hdulist),
            uncertainty=out_uncertainty,
            unit=unit,
            meta=hdulist[0].header,
            mask=mask,
            scaled=scaled,
            _basic_wcs=_create_wcs(hdulist),
        )
        [map_cube.extra_coords.add(*extra_coord) for extra_coord in extra_coords]
    return map_cube



import textwrap

import matplotlib.pyplot as plt
import numpy as np

import astropy.units as u
from astropy.coordinates import SkyCoord
from astropy.time import Time

from ndcube import NDCollection
from ndcube.meta import NDMeta
from sunpy import log as logger
from sunpy.coordinates import Helioprojective
from sunraster import SpectrogramCube as SpecCube
from sunraster import SpectrogramSequence as SpecSeq
from sunraster.meta import SlitSpectrographMetaABC
from sunraster.spectrogram import APPLY_EXPOSURE_TIME_ERROR

from irispy import utils
from irispy.utils.constants import SPECTRAL_BAND
from irispy.visualization import IRISPlotter, IRISSequencePlotter, set_axis_properties

__all__ = ["Collection", "SGMeta", "SpectrogramCube", "SpectrogramCubeSequence"]


class Collection(NDCollection):
    def __str__(self) -> str:
        return textwrap.dedent(
            f"""
            Collection
            ----------
            Cube keys: {tuple(self.keys())}
            Number of Cubes: {len(self)}
            Aligned dimensions: {self.aligned_dimensions}
            Aligned physical types: {self.aligned_axis_physical_types}
            """,
        )


class SpectrogramCube(SpecCube):
    """
    Class representing SpectrogramCube data described by a single WCS.

    Parameters
    ----------
    data: `numpy.ndarray`
        The array holding the actual data in this object.
    wcs: `ndcube.wcs.wcs.WCS`
        The WCS object containing the axes' information
    unit : `astropy.unit.Unit` or `str`
        Unit for the dataset. Strings that can be converted to a Unit are allowed.
    meta : dict-like object
        Additional meta information about the dataset. Must contain at least the
        following keys:
        - detector type: str, (FUV1, FUV2 or NUV)
        - OBSID: int
        - spectral window: str
    uncertainty : any type, optional
        Uncertainty in the dataset. Should have an attribute uncertainty_type
        that defines what kind of uncertainty is stored, for example "std"
        for standard deviation or "var" for variance. A metaclass defining
        such an interface is NDUncertainty - but isn't mandatory. If the uncertainty
        has no such attribute the uncertainty is stored as UnknownUncertainty.
        Defaults to None.
    mask : any type, optional
        Mask for the dataset. Masks should follow the numpy convention
        that valid data points are marked by False and invalid ones with True.
        Defaults to None.
    copy : `bool`, optional
        Indicates whether to save the arguments as copy. True copies every attribute
        before saving it while False tries to save every parameter as reference.
        Note however that it is not always possible to save the input as reference.
        Default is False.
    """

    def __init__(
        self,
        data,
        wcs,
        uncertainty,
        unit,
        meta,
        *,
        mask=None,
        copy=False,
    ) -> None:
        super().__init__(
            data,
            wcs,
            unit=unit,
            uncertainty=uncertainty,
            mask=mask,
            meta=meta,
            copy=copy,
        )

    def __getitem__(self, item):
        result = super().__getitem__(item)
        return SpectrogramCube(
            result.data,
            result.wcs,
            result.uncertainty,
            result.unit,
            result.meta,
            mask=result.mask,
        )

    def __repr__(self) -> str:
        return f"{object.__repr__(self)}\n{self!s}"

    def __str__(self) -> str:
        instance_start = None
        instance_end = None
        if self.global_coords and "time" in self.global_coords:
            instance_start = self.global_coords["time"].min().isot
            instance_end = self.global_coords["time"].max().isot
        elif self.extra_coords and self.axis_world_coords("time", wcs=self.extra_coords):
            instance_start = self.axis_world_coords("time", wcs=self.extra_coords)[0].min().isot
            instance_end = self.axis_world_coords("time", wcs=self.extra_coords)[0].max().isot
        return textwrap.dedent(
            f"""
            SpectrogramCube
            ---------------
            OBS ID:             {self.meta.get("OBSID")}
            OBS Description:    {self.meta.get("OBS_DESC")}
            OBS period:         {self.meta.get("STARTOBS")} -- {self.meta.get("ENDOBS")}
            Spectrogram period: {instance_start} -- {instance_end}
            Data shape:         {self.shape}
            Axis Types:         {self.array_axis_physical_types}
            Roll:               {self.meta.get("SAT_ROT")}
            """,
        )

    def plot(self, *args, **kwargs):
        cmap = kwargs.get("cmap")
        if not cmap:
            try:
                cmap = plt.get_cmap(name=f"irissji{int(self.meta.detector[:3])}")
            except Exception as e:  # NOQA: BLE001
                logger.debug(e)
                cmap = "viridis"
        kwargs["cmap"] = cmap
        if len(self.shape) == 1:
            kwargs.pop("cmap")
        ax = IRISPlotter(ndcube=self).plot(*args, **kwargs)
        set_axis_properties(ax)
        return ax

    def convert_to(self, new_unit_type, time_obs=None, response_version=4):
        """
        Converts data, unit and uncertainty attributes to new unit type.

        Takes into consideration also the observation time and response version.

        The presence or absence of the exposure time correction is
        preserved in the conversions.

        Parameters
        ----------
        new_unit_type : `str`
            Unit type to convert data to. Three values are accepted:
            "DN": Relevant IRIS data number based on detector type.
            "photons": photon counts
            "radiance": Perorms radiometric calibration conversion.
        time_obs : `astropy.time.Time`, optional
            Observation times of the datapoints.
            Must be in the format of, e.g.,
            ``time_obs=Time('2013-09-03')``,
            The argument time_obs is ignored for versions 1 and 2.
        response_version : `int`, optional
            Version number of effective area file to be used, by default = 6.

        Returns
        -------
        `SpectrogramCube`
            New SpectrogramCube in new units.
        """
        detector_type = utils.get_detector_type(self.meta)
        if new_unit_type == "radiance" or self.unit.is_equivalent(utils.RADIANCE_UNIT):
            # Get spectral dispersion per pixel.
            spectral_wcs_index = np.where(np.array(self.wcs.wcs.ctype) == "WAVE")[0][0]
            spectral_dispersion_per_pixel = (
                self.wcs.wcs.cdelt[spectral_wcs_index] * self.wcs.wcs.cunit[spectral_wcs_index]
            )
            # Get solid angle from slit width for a pixel.
            lat_wcs_index = ["HPLT" in c for c in self.wcs.wcs.ctype]
            lat_wcs_index = np.arange(len(self.wcs.wcs.ctype))[lat_wcs_index]
            lat_wcs_index = lat_wcs_index[0]
            solid_angle = self.wcs.wcs.cdelt[lat_wcs_index] * self.wcs.wcs.cunit[lat_wcs_index] * utils.SLIT_WIDTH
            # Get wavelength for each pixel.
            obs_wavelength = self.axis_world_coords(2)

        if new_unit_type in ["DN", "photons"]:
            if self.unit.is_equivalent(utils.RADIANCE_UNIT):
                # Convert from radiance to counts/s
                new_data_quantities = utils.convert_or_undo_photons_per_sec_to_radiance(
                    (self.data * self.unit, self.uncertainty.array * self.unit),
                    time_obs,
                    response_version,
                    obs_wavelength,
                    detector_type,
                    spectral_dispersion_per_pixel,
                    solid_angle,
                    undo=True,
                )
                new_data = new_data_quantities[0].value
                new_uncertainty = new_data_quantities[1].value
                new_unit = new_data_quantities[0].unit
                self = SpectrogramCube(  # NOQA: PLW0642 Reassigned `self` variable in instance method
                    new_data,
                    self.wcs,
                    new_uncertainty,
                    new_unit,
                    self.meta,
                    mask=self.mask,
                )
                self._extra_coords = self.extra_coords
            new_unit = utils.DN_UNIT[detector_type] if new_unit_type == "DN" else u.photon
            new_data_arrays, new_unit = utils.convert_between_dn_and_photons(
                (self.data, self.uncertainty.array),
                self.unit,
                new_unit,
            )
            new_data = new_data_arrays[0]
            new_uncertainty = new_data_arrays[1]
        elif new_unit_type == "radiance":
            if self.unit.is_equivalent(utils.RADIANCE_UNIT):
                new_data = self.data
                new_uncertainty = self.uncertainty
                new_unit = self.unit
            else:
                # Ensure spectrogram is in units of counts/s.
                cube = self.convert_to("photons")
                try:
                    cube = cube.apply_exposure_time_correction()
                except ValueError(APPLY_EXPOSURE_TIME_ERROR):
                    pass
                # Convert to radiance units.
                new_data_quantities = utils.convert_or_undo_photons_per_sec_to_radiance(
                    (cube.data * cube.unit, cube.uncertainty.array * cube.unit),
                    time_obs,
                    response_version,
                    obs_wavelength,
                    detector_type,
                    spectral_dispersion_per_pixel,
                    solid_angle,
                )
                new_data = new_data_quantities[0].value
                new_uncertainty = new_data_quantities[1].value
                new_unit = new_data_quantities[0].unit
        else:
            msg = "Input unit type not recognized."
            raise ValueError(msg)
        new_cube = SpectrogramCube(
            new_data,
            self.wcs,
            new_uncertainty,
            new_unit,
            self.meta,
            mask=self.mask,
        )
        new_cube._extra_coords = self.extra_coords
        return new_cube


class SpectrogramCubeSequence(SpecSeq):
    """
    Class for holding, slicing and plotting IRIS spectrogram data.

    This class contains all the functionality of its super class with
    some additional functionalities.

    Parameters
    ----------
    data_list: `list`
        List of `SpectrogramCube` objects from the same spectral window and OBS ID.
    meta: `dict` or header object, optional
        Metadata associated with the sequence.
    common_axis: `int`, optional
        The axis of the NDCubes corresponding to time.
    """

    def __init__(self, data_list, meta=None, common_axis=0) -> None:
        # Check that all spectrograms are from same spectral window and OBS ID.
        if len(np.unique([cube.meta["OBSID"] for cube in data_list])) != 1:
            msg = "Constituent SpectrogramCube objects must have same value of 'OBSID' in its meta."
            raise ValueError(msg)
        super().__init__(data_list, meta=meta, common_axis=common_axis)

    def __str__(self) -> str:
        # Overload it get the class name in the string
        return super().__str__()

    def plot(self, *args, **kwargs):
        cmap = kwargs.get("cmap")
        if not cmap:
            try:
                cmap = plt.get_cmap(name=f"irissji{int(self.meta.detector[:3])}")
            except Exception as e:  # NOQA: BLE001
                logger.debug(e)
                cmap = "viridis"
        kwargs["cmap"] = cmap
        if len(self.shape) == 1:
            kwargs.pop("cmap")
        ax = IRISSequencePlotter(ndcube=self).plot(*args, **kwargs)
        set_axis_properties(ax)
        return ax

    def convert_to(self, new_unit_type, *, copy=False):
        """
        Converts data, uncertainty and unit of each spectrogram in sequence to
        new unit.

        Parameters
        ----------
        new_unit_type: `str`
            Unit type to convert data to.  Three values are accepted:
            "DN": Relevant IRIS data number based on detector type.
            "photons": photon counts
            "radiance": Performs radiometric calibration conversion.
        copy: `bool`
            If True a new instance with the converted data values is return.
            If False, the current instance is overwritten.
            Default=False
        """
        converted_data_list = [cube.convert_to(new_unit_type) for cube in self.data]
        if copy is True:
            return SpectrogramCubeSequence(converted_data_list, meta=self.meta, common_axis=self._common_axis)
        self.data = converted_data_list
        return None


class SGMeta(NDMeta, SlitSpectrographMetaABC):
    def __init__(self, header, spectral_window, **kwargs) -> None:
        super().__init__(header, **kwargs)
        spectral_windows = np.array([self[f"TDESC{i}"] for i in range(1, self["NWIN"] + 1)])
        window_mask = np.array([spectral_window in window for window in spectral_windows])
        if window_mask.sum() < 1:
            msg = (
                "Spectral window not found. "
                f"Input spectral window: {spectral_window}; "
                f"Spectral windows in header: {spectral_windows}"
            )
            raise ValueError(
                msg,
            )
        if window_mask.sum() > 1:
            msg = (
                "Spectral window must be unique. "
                f"Input spectral window: {spectral_window}; "
                f"Ambiguous spectral windows in header: {spectral_windows[window_mask]}"
            )
            raise ValueError(
                msg,
            )
        self._iwin = np.arange(len(spectral_windows))[window_mask][0] + 1
        self._fits_header = header

    def __str__(self) -> str:
        return textwrap.dedent(
            f"""
                SGMeta
                ------
                Observatory:     {self.observatory}
                Instrument:      {self.instrument}
                Detector:        {self.detector}
                Spectral Window: {self.spectral_window}
                Spectral Range:  {self.spectral_range}
                Spectral Band:   {self.spectral_band}
                Dimensions:      {self.data_shape}
                Date:            {self.date_reference}
                OBS ID:          {self.observing_mode_id}
                OBS Description: {self.observing_mode_description}
                """,
        )

    def __repr__(self) -> str:
        return f"{object.__repr__(self)}\n{self!s}"

    def _construct_time(self, key):
        val = self.get(key)
        if val is not None:
            val = Time(val, format="fits", scale="utc")
        return val

    @property
    def fits_header(self):
        return self._fits_header

    @property
    def spectral_window(self):
        return self.get(f"TDESC{self._iwin}")

    @property
    def detector(self):
        return self.get(f"TDET{self._iwin}")

    @property
    def instrument(self):
        return self.get("INSTRUME")

    @property
    def observatory(self):
        return self.get("TELESCOP")

    @property
    def processing_level(self):
        return self.get("DATA_LEV")

    @property
    def distance_to_sun(self):
        return self.get("DSUN_OBS") * u.m

    @property
    def date_reference(self):
        return self._construct_time("DATE_OBS")

    @property
    def date_start(self):
        return self.date_reference

    @property
    def date_end(self):
        return self._construct_time("DATE_END")

    @property
    def observing_mode_id(self):
        return int(self.get("OBSID"))

    # ---------- IRIS-specific metadata properties ----------
    @property
    def dimensions(self):
        return self.shape.tolist()

    @property
    def observing_mode_description(self):
        return self.get("OBS_DESC")

    @property
    def observing_campaign_start(self):
        """
        Start time of observing campaign.
        """
        return self._construct_time("STARTOBS")

    @property
    def observing_campaign_end(self):
        """
        End time of observing mode.
        """
        return self._construct_time("ENDOBS")

    @property
    def observation_includes_saa(self):
        """
        Whether IRIS passed through SAA during observations.
        """
        return bool(self.get("SAA"))

    @property
    def satellite_rotation(self):
        """
        Satellite roll from solar north.
        """
        return self.get("SAT_ROT") * u.deg

    @property
    def exposure_control_triggers_in_observation(self):
        """
        Number of times automatic exposure control triggered during observing
        campaign.
        """
        return self.get("AECNOBS")

    @property
    def exposure_control_triggers_in_raster(self):
        """
        Number of times automatic exposure control was triggered during this
        raster.
        """
        return self.get("AECNRAS")

    @property
    def number_raster_positions(self):
        """
        Number of positions in raster.
        """
        self.get("NRASTERP")

    @property
    def spectral_range(self):
        """
        The spectral range of the spectral window.
        """
        return [self.get(f"TWMIN{self._iwin}"), self.get(f"TWMAX{self._iwin}")] * u.AA

    @property
    def spectral_band(self):
        """
        The spectral band of the spectral window.
        """
        return SPECTRAL_BAND[self.spectral_window]

    @property
    def raster_fov_width_y(self):
        """
        Width of the field of view of the raster in the Y (slit) direction.
        """
        return self.get("FOVY") * u.arcsec

    @property
    def raster_fov_width_x(self):
        """
        Width of the field of view of the raster in the X (rastering)
        direction.
        """
        return self.get("FOVX") * u.arcsec

    @property
    def fov_center(self):
        """
        Location of the center of the field of view.
        """
        return SkyCoord(
            Tx=self.get("XCEN"),
            Ty=self.get("YCEN"),
            unit=u.arcsec,
            frame=Helioprojective,
        )

    @property
    def automatic_exposure_control_enabled(self):
        return bool(self.get("IAECFLAG"))

    @property
    def tracking_mode_enabled(self):
        return bool(self.get("TR_MODE"))

    @property
    def observatory_at_high_latitude(self):
        """
        Whether IRIS passed through high Earth latitude during observations.
        """
        return bool(self.get("HLZ"))

    @property
    def spatial_summing_factor(self):
        """
        Number of pixels summed together in the spatial (Y/slit) direction.
        """
        return self.get("SUMSPAT")

    @property
    def spectral_summing_factor(self):
        """
        Number of pixels summed together in the spectral direction.
        """
        if "fuv" in self.detector.lower():
            return self.get("SUMSPTRF")
        return self.get("SUMSPTRN")



import os

import numpy as np
import pooch
import pytest

from astropy.io import fits

from sunpy.time import parse_time

from irispy.data.test import get_test_filepath
from irispy.io.sji import read_sji_lvl2
from irispy.utils import get_iris_response


@pytest.fixture
def remote_raster_scanning_tar():
    return pooch.retrieve(
        "https://github.com/LM-SAL/irispy-lmsal-test-data/raw/refs/heads/main/iris_l2_20250613_123658_3620107423/iris_l2_20250613_123658_3620107423_raster.tar.gz",
        known_hash="756ca99cbdfafca2a97c3e357a9e8ab1bc897bca6991f6e0fa42ac2717d5b05a",
    )


@pytest.fixture
def raster_file():
    return get_test_filepath("iris_l2_20210905_001833_3620258102_raster_t000_r00000_test.fits")


@pytest.fixture
def sji_1330_file():
    return get_test_filepath("iris_l2_20210905_001833_3620258102_SJI_1330_t000_test.fits")


@pytest.fixture
def sji_1400_file():
    return get_test_filepath("iris_l2_20210905_001833_3620258102_SJI_1400_t000_test.fits")


@pytest.fixture
def sji_2796_file():
    return get_test_filepath("iris_l2_20210905_001833_3620258102_SJI_2796_t000_test.fits")


@pytest.fixture
def sji_2832_file():
    return get_test_filepath("iris_l2_20210905_001833_3620258102_SJI_2832_t000_test.fits")


@pytest.fixture
def sjicube_1330():
    return read_sji_lvl2(get_test_filepath("iris_l2_20210905_001833_3620258102_SJI_1330_t000_test.fits"))


@pytest.fixture
def sjicube_1400():
    return read_sji_lvl2(get_test_filepath("iris_l2_20210905_001833_3620258102_SJI_1400_t000_test.fits"))


@pytest.fixture
def sjicube_2796():
    return read_sji_lvl2(get_test_filepath("iris_l2_20210905_001833_3620258102_SJI_2796_t000_test.fits"))


@pytest.fixture
def sjicube_2832():
    return read_sji_lvl2(get_test_filepath("iris_l2_20210905_001833_3620258102_SJI_2832_t000_test.fits"))


@pytest.fixture
def iris_response_v1():
    return get_iris_response(time_obs=parse_time("2013-09-03"), response_version=1)


@pytest.fixture
def iris_response_v2():
    return get_iris_response(time_obs=parse_time("2013-09-03"), response_version=2)


@pytest.fixture
def iris_response_v3():
    return get_iris_response(time_obs=parse_time("2013-09-03"), response_version=3)


@pytest.fixture
def iris_response_v4():
    return get_iris_response(time_obs=parse_time("2013-09-03"), response_version=4)


@pytest.fixture
def iris_response_v5():
    return get_iris_response(time_obs=parse_time("2013-09-03"), response_version=5)


@pytest.fixture
def iris_response_v6():
    return get_iris_response(time_obs=parse_time("2013-09-03"), response_version=6)


@pytest.fixture
def filelist():
    return [
        get_test_filepath("iris_l2_20210905_001833_3620258102_SJI_1330_t000_test.fits"),
        get_test_filepath("iris_l2_20210905_001833_3620258102_SJI_1400_t000_test.fits"),
        get_test_filepath("iris_l2_20210905_001833_3620258102_SJI_2796_t000_test.fits"),
        get_test_filepath("iris_l2_20210905_001833_3620258102_SJI_2832_t000_test.fits"),
    ]


@pytest.fixture(scope="session")
def fake_long_obs(tmp_path_factory):
    header = fits.getheader(get_test_filepath("iris_l2_20210905_001833_3620258102_SJI_2832_t000_test.fits"))
    header["STARTOBS"] = "2017-05-02T05:25:51.000"
    header["ENDOBS"] = "2017-05-02T08:25:51.000"
    header["NAXIS3"] = 100
    if header["CUNIT3"] == "seconds":
        header["CUNIT3"] = "s"
    rng = np.random.default_rng(12345)
    data = rng.random((header["NAXIS3"], header["NAXIS2"], header["NAXIS1"]))
    temp_dir = tmp_path_factory.mktemp("IRIS")
    hdu = fits.PrimaryHDU(data=data, header=header, do_not_scale_image_data=True, scale_back=True)
    fits_file = os.fspath(temp_dir.joinpath("iris_l2_20210905_001833_3620258102_SJI_2832_t000_test.fits"))
    hdu.writeto(fits_file)
    return [fits_file]



"""
Short script I used to create the test FITS files in this folder.
"""


def compress(files: list) -> None:
    from scipy.ndimage import zoom  # NOQA: PLC0415

    from astropy.io import fits  # NOQA: PLC0415

    for file in files:
        hdus = fits.open(file)
        for hdu in hdus:
            hdu.verify("fix")
            # TODO: why is this here?
            if "CADPL_DV" in hdu.header:
                print(hdu.header["CADPL_DV"])  # NOQA: T201
                del hdu.header["CADPL_DV"]
            # TODO: why is this here?
            if "CADEX_DV" in hdu.header:
                print(hdu.header["CADEX_DV"])  # NOQA: T201
                del hdu.header["CADEX_DV"]
            if hdu.data is None:
                continue
            # Can't pop out the array, resizing can cause issues
            # So I remove the data and move on.
            if hdu.data.ndim == 1:
                hdu.data = None
                continue
            if hdu.data.ndim == 2:
                factor = (0.1, 1)
            elif hdu.data.ndim == 3:
                factor = (0.1, 0.1, 0.1)
            hdu.data = zoom(hdu.data, factor, order=0)
            if hdu.data.ndim == 2:
                hdu.header["NAXIS1"] = hdu.data.shape[1]
                hdu.header["NAXIS2"] = hdu.data.shape[0]
            elif hdu.data.ndim == 3:
                hdu.header["NAXIS1"] = hdu.data.shape[2]
                hdu.header["NAXIS2"] = hdu.data.shape[1]
                hdu.header["NAXIS3"] = hdu.data.shape[0]
        hdus.writeto(f"{file.split('.fits')[0]}_test.fits", overwrite=True)



"""
=================
Cropping IRIS SJI
=================

In this example we will show crop a IRIS dataset and the particularity of the crop operation.
"""

import matplotlib.pyplot as plt
import pooch

import astropy.units as u
from astropy.coordinates import SkyCoord
from astropy.wcs.utils import wcs_to_celestial_frame

from irispy.io import read_files
from irispy.obsid import ObsID

###############################################################################
# We start by downloading the data from the IRIS archive.
#
# In this case, we will use ``pooch`` to keep this example self-contained
# but using your browser will also work.

sji_filename = pooch.retrieve(
    "http://www.lmsal.com/solarsoft/irisa/data/level2_compressed/2014/09/19/20140919_051712_3860608353/iris_l2_20140919_051712_3860608353_SJI_2832_t000.fits.gz",
    known_hash="02a1cdbe2014e24b04ff782dcfdbaf553c5a03404813888ddea8c50a9d6b2630",
)

###############################################################################
# We will now open the Slit-Jaw Imager (SJI) file we just downloaded.

sji_2832 = read_files(sji_filename)
# Printing will give us an overview of the file.
print(sji_2832)
# ``.meta`` contains the entire FITS header from the primary HDU.
# Since it is very long, we won't actually print it here.

###############################################################################
# Can't remember what is OBSID 3860608353?
#
# **irispy-lmsal** has an utility function that will provide more information.

print(ObsID(sji_2832.meta["OBSID"]))

###############################################################################
# Now, we will plot the SJI. By default, **irispy-lmsal** will
# color the spatial axes.

# This is an animation
sji_2832.plot()

###############################################################################
# We also have the option of going directly to an individual scan.

sji_45 = sji_2832[45]
print(sji_45)

###############################################################################
# We need to get the coordinate frame for the IRIS data.
# While this is stored in the WCS, getting a coordinate frame is a little more involved.
# We will use this to do a cutout later on.

sji_frame = wcs_to_celestial_frame(sji_45.basic_wcs)
bbox = [
    SkyCoord(-750 * u.arcsec, 90 * u.arcsec, frame=sji_frame),
    SkyCoord(-750 * u.arcsec, 95 * u.arcsec, frame=sji_frame),
    SkyCoord(-700 * u.arcsec, 90 * u.arcsec, frame=sji_frame),
    SkyCoord(-700 * u.arcsec, 95 * u.arcsec, frame=sji_frame),
]

###############################################################################
# This dataset has a peculiarity: the observation has a 45 degree roll.
# The image does not have a 45 degree rotation because plotting shows the data
# in the way they are written in the file.
# We will a coordinate grid to make this clear.
#
# You can also change the axis labels and ticks if you so desire.
# `WCSAxes provides us an API we can use. <https://docs.astropy.org/en/stable/visualization/wcsaxes/index.html>`__
#
# Now, let us cut out the top sunspot.
#
# We need to specify the corners for the cut (note the coordinate order is
# the same as the plotted image). Be aware that crop works in the default N/S frame,
# so it will crop along those axes where as the data is rotated.
# You will also need to create a proper bounding box, with 4 corners.
#
# ``crop`` will return you the smallest bounding box which contains those 4 points
# which we can see when we overlay the points we give it.
# So despite the bounding box being the incorrect location, it returns the cutout we want.

sji_cutout = sji_45.crop(*bbox)

plt.figure()
ax = sji_cutout.plot()
# Plot each corner of the box
[ax.plot_coord(coord, "o") for coord in bbox]
ax.coords.grid()

plt.show()



import copy

import numpy as np

from astropy.io import fits

from irispy.io.spectrograph import read_spectrograph_lvl2


def test_fits_data_comparison(raster_file):
    """
    Make sure the data is the same in pyfits and irispy.
    """
    iris_l2_test_raster = read_spectrograph_lvl2(raster_file)
    with fits.open(raster_file) as hdulist:
        spectral_window1 = hdulist[0].header["TDESC1"]
        spectral_window2 = hdulist[0].header["TDESC2"]
        spectral_window3 = hdulist[0].header["TDESC3"]
        data1 = copy.deepcopy(hdulist[1].data)
        data2 = copy.deepcopy(hdulist[2].data)
        data3 = copy.deepcopy(hdulist[3].data)
        np.testing.assert_array_almost_equal(iris_l2_test_raster[spectral_window1].data[0].data, data1)
        np.testing.assert_array_almost_equal(iris_l2_test_raster[spectral_window2].data[0].data, data2)
        np.testing.assert_array_almost_equal(iris_l2_test_raster[spectral_window3].data[0].data, data3)



import tarfile
from pathlib import Path

from astropy.io import fits

from ndcube import NDCollection
from sunpy import log as logger

from irispy.io.sji import read_sji_lvl2
from irispy.io.spectrograph import read_spectrograph_lvl2

__all__ = ["fitsinfo", "read_files"]


def _get_simple_metadata(file):
    """
    Get simple metadata from a FITS file.

    Parameters
    ----------
    file : `pathlib.Path`
        The FITS file to read.

    Returns
    -------
    `tuple`
        A tuple containing the instrument name and description.
    """
    if not file.name.endswith(".fits") and not file.name.endswith(".fits.gz"):
        return "", ""
    instrume = fits.getval(file, "INSTRUME")
    describe = fits.getval(file, "TDESC1")
    return instrume, describe


def _extract_tarfile(filenames):
    """
    Extracts a tar file to the same location as the tar file.

    Parameters
    ----------
    filenames : `list of str`
        The filenames of the tar files to extract.
    """
    expanded_files = []
    for fname in filenames:
        filename = Path(fname)
        if tarfile.is_tarfile(filename):
            extract_dir = filename.with_suffix("").with_suffix("")  # removes .tar.gz or .tar
            extract_dir.mkdir(parents=True, exist_ok=True)
            with tarfile.open(filename, "r") as tar:
                tar.extractall(extract_dir, filter="data")
                expanded_files.extend([extract_dir / member.name for member in tar.getmembers() if member.isfile()])
        else:
            expanded_files.append(filename)
    return expanded_files


def fitsinfo(filename):
    """
    Prints information about the extension of a raster or SJI level 2 data
    file.

    Parameters
    ----------
    filename : str
        Filename to load.
    """
    with fits.open(filename) as hdulist:
        hdulist.info()
        hdr = hdulist[0].header
        msg = f"Observation description: {hdr['OBS_DESC']}"
        logger.info(msg)
        modifier = ""
        for i in range(hdr["NWIN"]):
            msg = f"Extension No. {i + 1} stores data and header of {hdr[f'TDESC{i + 1}']}: "
            logger.info(msg)
            if "SJI" not in hdr[f"TDET{i + 1}"]:
                modifier = f" ({hdr[f'TDET{i + 1}'][:3]})"
            msg = f"{hdr[f'TWMIN{i + 1}']:.2f} - {hdr[f'TWMAX{i + 1}']:.2f} AA{modifier}"
            logger.info(msg)


def read_files(filenames, *, spectral_windows=None, uncertainty=False, memmap=False, allow_errors=False, **kwargs):
    """
    A wrapper function to read any number of raster, SJI or IRIS-aligned AIA
    data files.

    The goal is be able to download an entire IRIS observation and read it
    in one go, without having to worry about the type of file.

    Parameters
    ----------
    filename : `list of `str`, `str`, `pathlib.Path`
        Filename(s) to load.
    spectral_windows: iterable of `str` or `str`
        Spectral windows to extract from files. Default=None, implies, extract all
        spectral windows.
    uncertainty : `bool`, optional
        If `True` (not the default), will compute the uncertainty for the data (slower and
        uses more memory). If `memmap=True`, the uncertainty is never computed.
    memmap : `bool`, optional
        If `True` (not the default), will not load arrays into memory, and will only read from
        the file into memory when needed. This option is faster and uses a
        lot less memory. However, because FITS scaling is not done on-the-fly,
        the data units will be unscaled, not the usual data numbers (DN).
    allow_errors : `bool`, optional
        Will continue loading the files if one fails to load.
        Defaults to `False`.
    kwargs : `dict`, optional
        Additional keyword arguments to pass to the reader functions.

    Returns
    -------
    `NDCollection`
        With keys being the value of TDESC1, the values being the cube.
    """
    if isinstance(filenames, (str, Path)):
        filenames = [filenames]
    filenames = sorted(filenames)
    filenames = [Path(f) for f in filenames]
    returns = {}
    for filename in filenames:
        sdo_tarfile = bool(filename.name.endswith("SDO.tar.gz"))
        raster_tarfile = bool(filename.name.endswith("_raster.tar.gz"))
        instrume, describe = _get_simple_metadata(filename)
        logger.debug(f"Processing file: {filename} with instrume: {instrume}")
        try:
            if sdo_tarfile or instrume in ["IRIS", "SJI"] or instrume.startswith("AIA"):
                file = _extract_tarfile([filename]) if sdo_tarfile else [filename]
                for f in file:
                    instrume, describe = _get_simple_metadata(f)
                    returns[f"{describe}"] = read_sji_lvl2(f, memmap=memmap, uncertainty=uncertainty, **kwargs)
            elif raster_tarfile or instrume == "SPEC":
                file = _extract_tarfile([filename]) if raster_tarfile else [filename]
                instrume, describe = _get_simple_metadata(file[0])
                returns[f"{describe}"] = read_spectrograph_lvl2(
                    file, spectral_windows=spectral_windows, memmap=memmap, uncertainty=uncertainty, **kwargs
                )
            else:
                logger.warning(f"INSTRUME: {instrume} was not recognized and not loaded")
        except Exception as e:
            if allow_errors:
                logger.warning(f"File {filename} failed to load with {e}")
                continue
            raise
    return NDCollection(returns.items()) if len(returns) > 1 else next(iter(returns.values()))



"""
This module provides general utility functions for IRIS Responses.
"""

import datetime
from pathlib import Path

import numpy as np
import scipy
import scipy.io

import astropy.units as u
from astropy.time import Time
from astropy.units.quantity import Quantity

from sunpy.time import parse_time

from irispy.data import ROOTDIR

RESPONSE_VERSION_FILENAMES = {
    "1": "iris_sra_20130211.geny",
    "2": "iris_sra_20130715.geny",
    "3": "iris_sra_c_20150331.geny",
    "4": "iris_sra_c_20161022.geny",
    "5": "iris_sra_c_20191101.geny",
    "6": "iris_sra_c_20200223.geny",
}

__all__ = ["fit_iris_xput", "get_iris_response"]


def get_iris_response(
    time_obs=None,
    response_version=6,
):
    """
    Returns IRIS response structure.

    The pre-launch response is version 2.
    The latest version is 6 which is used by default.

    Parameters
    ----------
    time_obs: an `astropy.time.Time` object, as a kwarg, valid for version > 2 ,optional
        Observation times of the datapoints.
        Must be in the format of, e.g.,
        time_obs=parse_time('2013-09-03', format='utime'),
        which yields 1094169600.0 seconds in value.
        The argument time_obs is ignored for versions 1 and 2.
    response_version : `int`,optional
        Version number of effective area file to be used, by default is 6.

    Notes
    -----
    This routine does calculate time dependent effective areas using
    version 3 and above of the response functions as is done in the SSW version
    of this code. This code has been updated to calculate time-dependent
    effective areas.

    Returns
    -------
    `dict`
        Various parameters regarding IRIS response or effective area structure.
        Includes the following keys:
        date_obs : `astropy.time.Time`
        lambda : `astropy.units.Quantity`
        area_sg : `astropy.units.Quantity`
        name_sg : `str`
        dn2phot_sg : `tuple` of length 2
        area_sji : `astropy.units.Quantity`
        name_sji : `str`
        dn2phot_sji : `tuple` of length 4
        comment : `str`
        version : `int`
        version_date : `astropy.time.Time`
    """
    response_filename = RESPONSE_VERSION_FILENAMES[str(response_version)]
    path = Path(ROOTDIR) / response_filename
    raw_response_data = scipy.io.readsav(path)
    iris_response = {name: raw_response_data["p0"][name][0] for name in raw_response_data["p0"].dtype.names}
    # Convert some properties to more convenient types.
    iris_response["LAMBDA"] = Quantity(iris_response["LAMBDA"], unit=u.nm)
    iris_response["AREA_SG"] = Quantity(iris_response["AREA_SG"], unit=u.cm**2)
    iris_response["AREA_SJI"] = Quantity(iris_response["AREA_SJI"], unit=u.cm**2)
    iris_response["GEOM_AREA"] = Quantity(iris_response["GEOM_AREA"], unit=u.cm**2)
    iris_response["VERSION"] = int(iris_response["VERSION"])
    # Convert some properties not found in version below version 3 to
    # more convenient types.
    if iris_response["VERSION"] > 2:
        # If DATE_OBS has a value, convert to `astropy.time.Time`, else set to None.
        try:
            iris_response["DATE_OBS"] = parse_time(iris_response["DATE_OBS"], format="utime")
        except Exception:  # NOQA: BLE001
            iris_response["DATE_OBS"] = None
        time_obs = np.array([time_obs.value])
        iris_response["C_F_TIME"] = Time(iris_response["C_F_TIME"], format="utime")
        iris_response["C_F_LAMBDA"] = Quantity(iris_response["C_F_LAMBDA"], unit="nm")
        iris_response["C_N_TIME"] = Time(iris_response["C_N_TIME"], format="utime")
        iris_response["C_N_LAMBDA"] = Quantity(iris_response["C_N_LAMBDA"], unit="nm")
        iris_response["C_S_TIME"] = Time(iris_response["C_S_TIME"], format="utime")
        iris_response["ELEMENTS"]["DATE"] = parse_time(iris_response["ELEMENTS"]["DATE"].astype(str).tolist())
        iris_response["VERSION_DATE"] = parse_time(iris_response["VERSION_DATE"])
    if iris_response["VERSION"] <= 2:
        # Change DATE tag in data with version <= 2 to VERSION_DATE to
        # be consistent with more recent versions.
        iris_response["VERSION_DATE"] = Time(
            datetime.datetime(
                int(iris_response["DATE"][:4]),
                int(iris_response["DATE"][4:6]),
                int(iris_response["DATE"][6:8]),
                tzinfo=datetime.timezone.utc,
            ),
        )
        del iris_response["DATE"]
    if iris_response["VERSION"] > 2 and time_obs is not None:
        try:
            n_time_obs = len(time_obs)
        except Exception:  # NOQA: BLE001
            n_time_obs = 1
        iris_response["AREA_SG"] = np.zeros(iris_response["AREA_SG"].shape)
        iris_response["AREA_SJI"] = np.zeros(iris_response["AREA_SJI"].shape)
        # FUV SG effective areas
        lambran_fuv = np.array([[133.1, 135.9], [138.8, 140.8]])
        # Rough SG spectral ranges.  Setting effective area to 0 outside of these.
        shp_fuv = iris_response["COEFFS_FUV"].shape
        # Time-dependent response for shp_0[0] = 3 wavelengths
        iris_fit_fuv = np.zeros((n_time_obs, shp_fuv[0]))
        for j in range(shp_fuv[0]):
            iris_fit_fuv[:, j] = fit_iris_xput(
                time_obs,
                iris_response["C_F_TIME"],
                iris_response["COEFFS_FUV"][j, :, :],
            )
        # Interpolate onto lambda grid, separately for each of the two FUV CCD's.
        for j in range(2):
            w_fuv = np.logical_and(
                iris_response["LAMBDA"].value >= lambran_fuv[j, 0],
                iris_response["LAMBDA"].value <= lambran_fuv[j, 1],
            )
            for k in range(n_time_obs):
                interpol_fuv = scipy.interpolate.interp1d(
                    iris_response["C_F_LAMBDA"][j : j + 2],
                    np.squeeze(iris_fit_fuv[k, j : j + 2]),
                    fill_value="extrapolate",
                )
                iris_response["AREA_SG"][0, w_fuv] = interpol_fuv(iris_response["LAMBDA"][w_fuv])
        # 2. NUV SG effective areas
        lambran_nuv = np.array([278.2, 283.5])
        # Rough SG spectral ranges.  Setting effective area to 0 outside of these.
        shp_nuv = iris_response["COEFFS_NUV"].shape
        # Time-dependent response for shp_1[0] wavelengths
        iris_fit_nuv = np.zeros((n_time_obs, shp_nuv[0]))
        for j in range(shp_nuv[0]):
            iris_fit_nuv[:, j] = fit_iris_xput(
                time_obs,
                iris_response["C_N_TIME"],
                iris_response["COEFFS_NUV"][j, :, :],
            )
        # Interpolate onto lambda grid
        w_nuv = np.where(
            np.logical_and(
                iris_response["LAMBDA"].value >= lambran_nuv[0],
                iris_response["LAMBDA"].value <= lambran_nuv[1],
            ),
        )
        if iris_response["VERSION"] <= 3:
            for k in range(n_time_obs):
                interpol_nuv = scipy.interpolate.interp1d(
                    iris_response["C_N_LAMBDA"][:],
                    np.squeeze(iris_fit_nuv[k, :]),
                    fill_value="extrapolate",
                )
                iris_response["AREA_SG"][1, w_nuv] = interpol_nuv(iris_response["LAMBDA"][w_nuv])
        else:
            for k in range(n_time_obs):
                interpol_nuv = scipy.interpolate.CubicSpline(
                    iris_response["C_N_LAMBDA"][:],
                    np.squeeze(iris_fit_nuv[k, :]),
                    extrapolate=True,
                    bc_type="natural",
                    axis=0,
                )
                iris_response["AREA_SG"][1, w_nuv] = interpol_nuv(iris_response["LAMBDA"][w_nuv])
        # SJI effective areas
        if 2 < iris_response["VERSION"] < 4:  # Version 3 only
            for j in range(iris_response["COEFFS_SJI"].shape[0]):
                # Calculate pre-launch area from the individual elements
                prelaunch_area = iris_response["GEOM_AREA"]
                for k in range(len(iris_response["INDEX_EL_SJI"][j, :])):
                    index_values0 = iris_response["INDEX_EL_SJI"][j, k]
                    prelaunch_area = prelaunch_area * iris_response["ELEMENTS"][index_values0].trans
                # Time dependent response
                iris_fit_sji = fit_iris_xput(
                    time_obs,
                    iris_response["C_S_TIME"][j, :, :],
                    iris_response["COEFFS_SJI"][j, :, :],
                )
                # Time dependent profiles
                for k in range(n_time_obs):
                    iris_response["AREA_SJI"][j, :] = prelaunch_area * iris_fit_sji[k]
        else:  # Version 4 and above
            for nuv in range(2):
                # Calculate baseline SJI area curves
                area_sji = iris_response["GEOM_AREA"]
                for m in range(len(iris_response["INDEX_EL_SJI"][nuv * 2, :])):
                    index_values1 = iris_response["INDEX_EL_SJI"][nuv * 2 : nuv * 2 + 2, m]
                    area_sji = area_sji * iris_response["ELEMENTS"][index_values1].trans
                # Apply time dependent profile shape adjustment to FUV SJI
                if nuv == 0:
                    # FUV: apply FUV SG "slant", then normalize so that a weighted (2.4:1)
                    # sum at C II and Si IV gives constant response
                    weight = np.array([2.4, 1.0])  # Typical solar ratio CII : SiIV
                    wavelength = iris_response["C_F_LAMBDA"]
                    n_wavelength = len(wavelength)
                    wavelength = np.array(
                        [
                            wavelength[0].value,
                            (wavelength[n_wavelength - 2].value * 2.0 + wavelength[n_wavelength - 1].value) / 3.0,
                        ],
                    )  # 2 wavelengths in nm
                    # Calculate baseline SG area for scaling purposes
                    area_sg = iris_response["GEOM_AREA"]
                    for n in range(len(iris_response["INDEX_EL_SG"][nuv, :])):
                        index_values2 = iris_response["INDEX_EL_SG"][nuv, n]
                        area_sg = area_sg * iris_response["ELEMENTS"][index_values2].trans
                    # SG and SJI areas at wavelength
                    interpol_sg = scipy.interpolate.interp1d(
                        iris_response["LAMBDA"],
                        np.squeeze(area_sg),
                        fill_value="extrapolate",
                    )
                    area_sg2 = interpol_sg(wavelength)
                    area_sj2 = np.zeros((2, 2))
                    for n in range(2):
                        interpol_sji = scipy.interpolate.interp1d(
                            iris_response["LAMBDA"],
                            np.squeeze(area_sji[n]),
                            fill_value="extrapolate",
                        )
                        area_sj2[n, :] = interpol_sji(wavelength)
                    # Calculate the normalized slant function scal, apply to asji
                    for _ in range(n_time_obs):
                        # Best-estimate slant, i.e., eff.area @ wavelength / baseline SG @ wavelength
                        interpol_sg2 = scipy.interpolate.interp1d(
                            iris_response["LAMBDA"],
                            np.squeeze(iris_response["AREA_SG"][0, :]),
                            fill_value="extrapolate",
                        )
                        sca2 = interpol_sg2(wavelength) / area_sg2
                        # Normalize slant so that total(wei*asj2*sca2)/total(wei*asj2)=1
                        for m in range(2):
                            sca2n = sca2 * np.sum(weight * area_sj2[m, :]) / np.sum(weight * area_sj2[m, :] * sca2)
                            interpol_sca = scipy.interpolate.interp1d(
                                wavelength,
                                np.squeeze(sca2n),
                                fill_value="extrapolate",
                            )
                            sca1n = interpol_sca(iris_response["LAMBDA"])
                            sca1n = np.clip(sca1n, a_min=0, a_max=None)
                            iris_response["AREA_SJI"][m] = area_sji[m] * sca1n
                else:
                    # NUV: essentially same calculation as version = 3
                    for _ in range(n_time_obs):
                        iris_response["AREA_SJI"] = [Quantity(x, unit=u.cm**2) for x in iris_response["AREA_SJI"]]
                        area_sji = list(area_sji)
                        iris_response["AREA_SJI"][2:4] = area_sji[:]
            for j in range(4):
                # SJI specific time dependency
                iris_fit_sji = fit_iris_xput(
                    time_obs,
                    iris_response["C_S_TIME"][j, :, :],
                    iris_response["COEFFS_SJI"][j, :, :],
                )
                for k in range(n_time_obs):
                    iris_response["AREA_SJI"][j] = iris_response["AREA_SJI"][j] * iris_fit_sji[k]
    if not isinstance(iris_response["AREA_SG"], Quantity):
        iris_response["AREA_SG"] = Quantity(iris_response["AREA_SG"], unit=u.cm**2)
    if not isinstance(iris_response["AREA_SJI"], Quantity):
        iris_response["AREA_SJI"] = Quantity(iris_response["AREA_SJI"], unit=u.cm**2)

    return iris_response


def fit_iris_xput(time_obs, time_cal_coeffs, cal_coeffs):
    """
    To calculate the coefficients of best-fit time function for throughput,
    for which there are two modes:
    1. Perform fit: supply xput and single element ``cal_coeffs``.
    2. Apply fit: supply full ``cal_coeffs``.

    The procedure involved in this function is as follows:
    1. The time difference (in years) is computed from the ``time_obs`` and ``time_cal_coeffs``.
    2. A least-squares fit is performed to determine the best fit for the time-dependent
    effective areas given the time difference.

    Parameters
    ----------
    time_obs: `list`
        A list of observation times as `astropy.time.Time` objects.
    time_cal_coeffs: `astropy.time.Time` with
        Start and end times of intervals of constant ``cal_coeffs[i]``.
        These should be in "utime" format.
        a numpy array of floats (with exactly two columns)
    cal_coeffs: a numpy array of floats (with at least two columns)
        Coefficients of best-fit function.

    Returns
    -------
    `numpy.array`
        Yields the fit used to compute the effective area using the input times ``time_obs``.
    """
    time_obs = Time(parse_time(time_obs).utime, format="utime")
    if time_cal_coeffs.shape[1] != 2 or cal_coeffs.shape[1] < 2:
        # Raise ValueError as time coefficient have the wrong format.
        msg = "Incorrect number of elements either in time_cal_coeffs or in cal_coeffs."
        raise ValueError(msg)
    # Some time transformations.
    # Convert the time_cal_coeffs given in the .geny file into a ``astropy.time.Time``
    # object called t_cal_coeffs, so that the time differences will be in days...
    t_cal_coeffs = time_cal_coeffs.flatten().reshape(time_cal_coeffs.shape)
    # Exponent for transition between exp.decay intervals.
    transition_exp = 1.5
    # For loop for carrying out the least-squares fit and computation of fit output.
    fit_out = np.zeros(len(time_obs))
    for i, t in enumerate(time_obs):
        aux_cal_coeffs = np.zeros(2 * time_cal_coeffs.shape[0])
        fit_out = np.zeros(len(list(time_obs)))
        # Looking for the closest time in the calibration time intervals.
        # Differences are given in years before passing to the next stage.
        t_diff = Time(t, format="utime") - t_cal_coeffs
        t_diff = t_diff.flatten()
        # To convert to an array, quantities need to be dimensionless, hence dividing out the unit.
        t_diff = np.array([x.to(u.year).value for x in t_diff])
        idx = np.where(t_diff < 0)[0]
        if idx.size == 0:
            idx = 1
        else:
            idx = idx[0]
            if idx == 0:
                idx = 1
        # If the t_obs is between the calibration time intervals of a
        # calibration file (idx % !=0) then the aux_coeffs are given by an
        # exponential (coefficient and exponential value).
        # If the t_obs is between the end calibration time interval of
        # a calibration file (cal_file_t) and the beginning calibration
        # time interval of the next calibration file (cal_file_t+1)
        # (idx% 2 == 0) then, the aux_coeffs are given by 4 values
        # corresponding to a partial exponential obtained from
        # cal_file_t and a complementary exponential obtained from the
        # cal_file_t+1
        if idx % 2 != 0:  # I.e., if idx is not even...
            dtt_0 = 1.0
            exp_0 = np.exp(cal_coeffs[idx // 2, 2] * (t_diff[idx - 1]))
            aux_cal_coeffs[idx - 1 : idx + 1] = np.array([dtt_0, dtt_0 * exp_0])
        else:
            dtt_1 = (t_diff[idx - 1] / (t_diff[idx - 1] - t_diff[idx])) ** transition_exp
            dtt_0 = 1.0 - dtt_1
            exp_0 = np.exp(cal_coeffs[(idx // 2) - 1, 2] * (t_diff[idx - 2]))
            exp_1 = np.exp(cal_coeffs[idx // 2, 2] * (t_diff[idx]))
            aux_cal_coeffs[idx - 2 : idx + 2] = np.array([dtt_0, dtt_0 * exp_0, dtt_1, dtt_1 * exp_1])
        fit_out[i] = np.matmul(aux_cal_coeffs, cal_coeffs[:, :2].reshape(aux_cal_coeffs.shape[0]))
    return fit_out



import os
from pathlib import Path
from urllib.parse import urljoin

from sunpy import log
from sunpy.util.config import _is_writable_dir, get_and_create_sample_dir
from sunpy.util.parfive_helpers import Downloader

_BASE_URLS = (
    "https://github.com/sunpy/data/raw/main/irispy-lmsal/",
    "https://github.com/sunpy/sample-data/raw/master/irispy-lmsal/",
    "http://data.sunpy.org/irispy-lmsal/",
)
_SAMPLE_DATA = {
    "AIA_1700": "aia_20140919_060030_1700_image_lev1.fits",
    "RASTER": "iris_l2_20211001_060925_3683602040_raster.tar.gz",
    "SJI_1330": "iris_l2_20211001_060925_3683602040_SJI_1330_t000.fits.gz",
    "SJI_1400": "iris_l2_20211001_060925_3683602040_SJI_1400_t000.fits.gz",
    "SJI_2796": "iris_l2_20211001_060925_3683602040_SJI_2796_t000.fits.gz",
    "SJI_2832": "iris_l2_20211001_060925_3683602040_SJI_2832_t000.fits.gz",
}
_SAMPLE_FILES = {v: k for k, v in _SAMPLE_DATA.items()}


def _download_sample_data(base_url, sample_files, overwrite):
    """
    Downloads a list of files.

    Parameters
    ----------
    base_url : str
        Base URL for each file.
    sample_files : list of tuples
        List of tuples that are (URL_NAME, SAVE_NAME).
    overwrite : bool
        Will overwrite a file on disk if True.

    Returns
    -------
    `parfive.Results`
        Download results. Will behave like a list of files.
    """
    dl = Downloader(overwrite=overwrite, progress=True)
    for url_file_name, fname in sample_files:
        url = urljoin(base_url, url_file_name)
        dl.enqueue_file(url, filename=fname)
    return dl.download()


def _retry_sample_data(results, new_url_base):
    # In case we have a broken file on disk, overwrite it.
    dl = Downloader(overwrite=True, progress=True)
    for err in results.errors:
        file_name = err.url.split("/")[-1]
        log.debug(f"Failed to download {_SAMPLE_FILES[file_name]} from {err.url}: {err.exception}")
        # Update the url to a mirror and requeue the file.
        new_url = urljoin(new_url_base, file_name)
        log.debug(f"Attempting redownload of {_SAMPLE_FILES[file_name]} using {new_url}")
        dl.enqueue_file(new_url, filename=err.filepath_partial)
    extra_results = dl.download()
    # Make a new results object which contains all the successful downloads
    # from the previous results object and this retry, and all the errors from
    # this retry.
    new_results = results + extra_results
    new_results._errors = extra_results._errors
    return new_results


def _handle_final_errors(results):
    for err in results.errors:
        file_name = err.url.split("/")[-1]
        log.debug(
            f"Failed to download {_SAMPLE_FILES[file_name]} from {err.url}: {err.exception}",
        )
        log.error(
            f"Failed to download {_SAMPLE_FILES[file_name]} from all mirrors, the file will not be available.",
        )


def _get_sampledata_dir():
    # Workaround for tox only. This is not supported as a user option
    sampledata_dir = os.environ.get("SUNPY_SAMPLEDIR", "")
    if sampledata_dir:
        sampledata_dir = Path(sampledata_dir).expanduser().resolve()
        _is_writable_dir(sampledata_dir)
    else:
        # Creating the directory for sample files to be downloaded
        sampledata_dir = Path(get_and_create_sample_dir())
    return sampledata_dir


def _get_sample_files(filename_list, *, no_download=False, force_download=False):
    """
    Returns a list of disk locations corresponding to a list of filenames for
    sample data, downloading the sample data files as necessary.

    Parameters
    ----------
    filename_list : `list` of `str`
        List of filenames for sample data
    no_download : `bool`
        If ``True``, do not download any files, even if they are not present.
        Default is ``False``.
    force_download : `bool`
        If ``True``, download all files, and overwrite any existing ones.
        Default is ``False``.

    Returns
    -------
    `list` of `pathlib.Path`
        List of disk locations corresponding to the list of filenames.  An entry
        will be ``None`` if ``no_download == True`` and the file is not present.

    Raises
    ------
    RuntimeError
        Raised if any of the files cannot be downloaded from any of the mirrors.
    """
    sampledata_dir = _get_sampledata_dir()
    fullpaths = [sampledata_dir / fn for fn in filename_list]
    if no_download:
        fullpaths = [fp if fp.exists() else None for fp in fullpaths]
    else:
        to_download = zip(filename_list, fullpaths, strict=False)
        if not force_download:
            to_download = [(fn, fp) for fn, fp in to_download if not fp.exists()]
        if to_download:
            results = _download_sample_data(_BASE_URLS[0], to_download, overwrite=force_download)
            # Try the other mirrors for any download errors
            if results.errors:
                for next_url in _BASE_URLS[1:]:
                    results = _retry_sample_data(results, next_url)
                    if not results.errors:
                        break
                else:
                    _handle_final_errors(results)
                    raise RuntimeError
    return fullpaths



import pytest

import astropy.units as u

from irispy.obsid import ObsID

OBSIDS = [
    3400109162,
    3677508065,
    3880903651,
    4050607445,
]
INVALID_OBSID = [
    4643502010,
    4050607495,
    3880903650,
    3680903685,
    335987081297,
    40,
]
TEST_DATA = {}
TEST_DATA["exptime"] = [
    8 * u.s,
    8 * u.s,
    30 * u.s,
    4 * u.s,
]
TEST_DATA["raster_desc"] = [
    "Very large coarse 64-step raster",
    "Very large dense 96-step raster",
    "Small sit-and-stare",
    "Very large dense raster (tight timing)",
]
TEST_DATA["sjis"] = [
    "C II   Si IV   Mg II h/k",
    "C II   Si IV   Mg II h/k   Mg II w",
    "Si IV",
    "Si IV   Mg II h/k",
]
TEST_DATA["binning"] = [
    "Spatial x 2, Spectral x 2",
    "Spatial x 1, Spectral x 1",
    "Spatial x 2, Spectral x 8",
    "Spatial x 2, Spectral x 2",
]
TEST_DATA["fuv_binning"] = [
    "FUV binned same as NUV",
    "FUV spectrally rebinned x 4",
    "FUV spectrally rebinned x 8",
    "FUV spectrally rebinned x 4",
]
TEST_DATA["sji_cadence"] = [
    "SJI cadence default",
    "SJI cadence 0.5x faster",
    "SJI cadence default",
    "SJI cadence 10s",
]
TEST_DATA["linelist"] = [
    "Large linelist",
    "Flare linelist 1",
    "Full readout",
    "Small linelist",
]


@pytest.mark.parametrize(
    ("attr_name", "test_input", "expected_output"),
    [(name, obs, output[i]) for (name, output) in TEST_DATA.items() for i, obs in enumerate(OBSIDS)],
)
def test_attribute(attr_name, test_input, expected_output):
    assert ObsID(test_input)[attr_name] == expected_output


@pytest.mark.parametrize("test_input", [INVALID_OBSID])
def test_invalid_obsid(test_input):
    with pytest.raises(ValueError, match="Invalid OBS ID: must have 10 digits."):
        ObsID(test_input)



# Try to use setuptools_scm to get the current version; this is only used
# in development installations from the git repository.
from pathlib import Path

try:
    from setuptools_scm import get_version

    version = get_version(root=Path("..") / "..", relative_to=__file__)
except ImportError as e:
    msg = "setuptools_scm not installed"
    raise ImportError(msg) from e
except Exception as e:
    msg = f"setuptools_scm broken with {e}"
    raise ValueError(msg) from e



import numpy as np
import pytest

from irispy.io.utils import fitsinfo, read_files


def test_fitsinfo(capsys, raster_file, sji_1330_file, sji_1400_file, sji_2796_file, sji_2832_file):
    fitsinfo(raster_file)
    captured = capsys.readouterr()
    assert raster_file in captured.out

    fitsinfo(sji_1330_file)
    captured = capsys.readouterr()
    assert sji_1330_file in captured.out

    fitsinfo(sji_1400_file)
    captured = capsys.readouterr()
    assert sji_1400_file in captured.out

    fitsinfo(sji_2796_file)
    captured = capsys.readouterr()
    assert sji_2796_file in captured.out

    fitsinfo(sji_2832_file)
    captured = capsys.readouterr()
    assert sji_2832_file in captured.out


def test_read_files_with_mix(raster_file, sji_1330_file):
    returns = read_files([raster_file, sji_1330_file])
    assert len(returns) == 2


def test_read_files_raster(raster_file):
    # Simple test to ensure it does not error
    assert read_files(raster_file)
    assert read_files([raster_file])


def test_read_files_sji(sji_1330_file, sji_1400_file, sji_2796_file, sji_2832_file):
    # Simple test to ensure it does not error
    assert read_files(sji_1330_file)
    assert read_files(sji_1400_file)
    assert read_files(sji_2796_file)
    assert read_files(sji_2832_file)
    assert read_files([sji_2832_file])


def test_read_files_sji_more_than_one(sji_1330_file, sji_1400_file):
    returns = read_files([sji_1330_file, sji_1400_file])
    assert len(returns) == 2


@pytest.mark.remote_data
def test_read_files_raster_scanning(remote_raster_scanning_tar):
    returns = read_files(remote_raster_scanning_tar)
    assert len(returns) == 8  # spectral windows
    np.testing.assert_array_equal(
        returns["C II 1336"].shape, (29, 4, 388, 186)
    )  # 29 time steps, 4 steps, 388 spatial pixels, 186 spectral pixels
    np.testing.assert_array_equal(returns.aligned_dimensions, [29, 4, 388])



"""
======================================
Reproject IRIS SJI (rolled) to SDO/AIA
======================================

In this example we will show how to reproject a rolled IRIS dataset to SDO/AIA.

The IRIS team at LMSAL provides AIA data cubes which are coaligned to the IRIS FOV
for each observation from https://iris.lmsal.com/search/

Therefore this example is more a showcase of functionally.
"""

import matplotlib.pyplot as plt
import numpy as np
import pooch

import astropy.units as u
from astropy.coordinates import SkyCoord
from astropy.time import Time, TimeDelta
from astropy.wcs.utils import wcs_to_celestial_frame

import sunpy.map
from aiapy.calibrate import update_pointing
from aiapy.calibrate.util import get_pointing_table
from sunpy.net import Fido
from sunpy.net import attrs as a
from sunpy.visualization.drawing import extent

from irispy.io import read_files
from irispy.obsid import ObsID

###############################################################################
# We start with getting the data.
# This is done by downloading the data from the IRIS archive.
#
# In this case, we will use ``pooch`` as to keep this example self-contained
# but using your browser will also work.

sji_filename = pooch.retrieve(
    "http://www.lmsal.com/solarsoft/irisa/data/level2_compressed/2014/09/19/20140919_051712_3860608353/iris_l2_20140919_051712_3860608353_SJI_2832_t000.fits.gz",
    known_hash="02a1cdbe2014e24b04ff782dcfdbaf553c5a03404813888ddea8c50a9d6b2630",
)

###############################################################################
# We will now open the slit-jaw imager (SJI) file we just downloaded.

sji_2832 = read_files(sji_filename)
# Printing will give us an overview of the file.
print(sji_2832)
# ``.meta`` contains the entire FITS header from the primary HDU.
# Since it is very long, we won't actually print it here.

###############################################################################
# Can't remember what is OBSID 3860608353?
# **irispy-lmsal** has an utility function that will provide more information.

print(ObsID(sji_2832.meta["OBSID"]))

###############################################################################
# We also have the option of going directly to an individual scan.

sji_cut = sji_2832[45]
print(sji_cut)

###############################################################################
# We need to get the coordinate frame for the IRIS data.
# While this is stored in the WCS, getting a coordinate frame is a little more involved.

sji_frame = wcs_to_celestial_frame(sji_cut.basic_wcs)

###############################################################################
# This dataset has a peculiarity: the observation has a 45 degree roll.
# The image does not have a 45 degree rotation because plotting shows the data
# in the way they are written in the file.
# We will a coordinate grid to make this clear.
# You can also change the axis labels and ticks if you so desire.
# `WCSAxes provides us an API we can use. <https://docs.astropy.org/en/stable/visualization/wcsaxes/index.html>`__

plt.figure()
ax = sji_cut.plot()
plt.title(f"IRIS SJI {sji_2832.meta['TWAVE1']}", pad=20)
ax.coords.grid()

###############################################################################
# We will want to align the data to AIA.
# First we will want to pick a timestamp during the observation.
#
# Lets us now find the SJI observation where the time is closest to 06:00 on 2014-09-19.

(time_sji,) = sji_2832.axis_world_coords("time")
time_target = Time("2014-09-19T06:00:00.0")
time_index = np.abs(time_sji - time_target).argmin()
time_stamp = time_sji[time_index].isot
print(time_index, time_stamp)

###############################################################################
# The fact that it is rolled 45 degrees makes manual alignment tricky
# and will illustrate the usefulness of working with WCS.
# We will download an AIA 170 nm image from the VSO.
# Once we have acquired it, we will need to use **aiapy** to prep this image.

search_results = Fido.search(
    a.Time(time_stamp, Time(time_stamp) + TimeDelta(1 * u.minute), near=time_stamp),
    a.Instrument.aia,
    a.Wavelength(1700 * u.AA),
)
files = Fido.fetch(search_results, site="NSO")
aia_map = sunpy.map.Map(files[0])
pointing_table = get_pointing_table(
    source="JSOC",
    time_range=(Time(time_stamp) - TimeDelta(5 * 60 * u.minute), Time(time_stamp) + TimeDelta(1 * u.minute)),
)
aia_map = update_pointing(aia_map, pointing_table=pointing_table)
# You don't need to register AIA images unless you need them aligned to other AIA images.
# otherwise you are degrading the data as the affine transform is not perfect.
# But it is the last step to get a level 1.5 image.

###############################################################################
# Now let's plot the IRIS field of view on the AIA image.
#
# This IRIS data has no observer coordinate information
# **irispy-lmsal** will set this to be at Earth.
# This will allow us to transform from IRIS to any another observer.
#
# Using :meth:`~.draw_extent`, drawing regions is straightforward.

aia_bottom_left = SkyCoord(-850 * u.arcsec, -50 * u.arcsec, frame=aia_map.coordinate_frame)
aia_top_right = SkyCoord(-650 * u.arcsec, 150 * u.arcsec, frame=aia_map.coordinate_frame)
aia_sub = aia_map.submap(aia_bottom_left, top_right=aia_top_right)

fig = plt.figure()
ax = plt.subplot(projection=aia_sub)
aia_sub.plot()
extent(ax, sji_cut.basic_wcs)

###############################################################################
# We have a green square showing the region of the IRIS observation.
# To work with both IRIS and AIA data, it helps if the image axes are aligned,
# and for this we need to rotate one of them. We can either rotate SDO/AIA to the
# IRIS frame, or vice-versa.
#
# We will rotate the AIA data, using `sunpy.map.Map.reproject_to`.
# As `sunpy` does not support gWCS (yet), we have to use the basic WCS.

aia_reprojected = aia_sub.reproject_to(sji_cut.basic_wcs)

###############################################################################
# Now we can see the results.

fig = plt.figure()
ax1 = fig.add_subplot(1, 2, 1, projection=aia_reprojected.wcs)
aia_reprojected.plot(axes=ax1)
ax1.set_title("")

ax2 = fig.add_subplot(1, 2, 2, projection=sji_cut.basic_wcs)
sji_cut.plot(axes=ax2)

fig.tight_layout()

###############################################################################
# Finally, one way to visualize the alignment is to plot the AIA contours on the IRIS SJI image.

fig = plt.figure()

ax1 = fig.add_subplot(111, projection=sji_cut.wcs)
sji_cut.plot(axes=ax1)
aia_reprojected.draw_contours(levels=[500], colors=["red"], linewidths=2)
ax1.set_title("IRIS SJI with AIA contours")

plt.show()



from irispy.utils.wobble import wobble_movie


def test_wobble_movie(fake_long_obs, tmp_path):
    movies = wobble_movie(fake_long_obs, outdir=tmp_path)
    assert movies != []
    movies = wobble_movie(fake_long_obs, outdir=tmp_path, trim=True)
    assert movies != []



import numpy as np
import pytest

from astropy import units as u
from astropy.time import Time
from astropy.wcs import WCS

from ndcube.meta import NDMeta

from irispy import SJICube, utils

TIMES = Time(["2014-12-11T19:39:00.48", "2014-12-11T19:43:07.6"])
EXTRA_COORDS = [("time", 0, TIMES)]


@pytest.fixture
def cube():
    data = np.array(
        [
            [[1, 2, 3, 4], [2, 4, 5, 3], [0, 1, 2, 3]],
            [[2, 4, 5, 1], [10, 5, 2, 2], [10, 3, 3, 0]],
        ],
    )
    exposure_times = 2 * np.ones((2), float) * u.s
    uncertainty = np.sqrt(data)
    header = {
        "CTYPE1": "HPLN-TAN",
        "CUNIT1": "arcsec",
        "CDELT1": 0.4,
        "CRPIX1": 0,
        "CRVAL1": 0,
        "NAXIS1": 4,
        "CTYPE2": "HPLT-TAN",
        "CUNIT2": "arcsec",
        "CDELT2": 0.5,
        "CRPIX2": 0,
        "CRVAL2": 0,
        "NAXIS2": 3,
        "CTYPE3": "Time    ",
        "CUNIT3": "s",
        "CDELT3": 0.3,
        "CRPIX3": 0,
        "CRVAL3": 0,
        "NAXIS3": 2,
    }
    wcs = WCS(header=header, naxis=3)
    cube = SJICube(
        data,
        wcs,
        uncertainty=uncertainty,
        mask=data >= 0,
        unit=utils.DN_UNIT["SJI"],
        scaled=True,
        meta=NDMeta({"exposure time": exposure_times}, axes={"exposure time": 0}, data_shape=data.shape),
    )
    cube.extra_coords.add(*EXTRA_COORDS[0])
    return cube


@pytest.fixture
def cube_2d():
    data_2d = np.array([[1, 2, 3, 4], [2, 4, 5, 3]])
    uncertainty_2d = np.sqrt(data_2d)
    header_2d = {
        "CTYPE1": "HPLN-TAN",
        "CUNIT1": "arcsec",
        "CDELT1": 0.4,
        "CRPIX1": 0,
        "CRVAL1": 0,
        "NAXIS1": 4,
        "CTYPE2": "HPLT-TAN",
        "CUNIT2": "arcsec",
        "CDELT2": 0.5,
        "CRPIX2": 0,
        "CRVAL2": 0,
        "NAXIS2": 3,
    }
    exposure_times = 2 * np.ones((2), float) * u.s
    wcs_2d = WCS(header=header_2d, naxis=2)
    cube_2d = SJICube(
        data_2d,
        wcs_2d,
        uncertainty=uncertainty_2d,
        mask=data_2d >= 0,
        unit=utils.DN_UNIT["SJI"],
        scaled=True,
        meta=NDMeta({"exposure time": exposure_times}, axes={"exposure time": 0}, data_shape=data_2d.shape),
    )
    cube_2d.extra_coords.add(*EXTRA_COORDS[0])
    return cube_2d


@pytest.fixture
def cube_1d():
    header_1d = {
        "CTYPE1": "Time    ",
        "CUNIT1": "s",
        "CDELT1": 0.4,
        "CRPIX1": 0,
        "CRVAL1": 0,
        "NAXIS1": 2,
    }
    exposure_times = 2 * np.ones((2), float) * u.s
    wcs_1d = WCS(header=header_1d, naxis=1)
    data_1d = np.array([1, 2])
    cube_1d = SJICube(
        data_1d,
        wcs_1d,
        uncertainty=np.sqrt(np.array([1, 2])),
        mask=data_1d >= 0,
        unit=utils.DN_UNIT["SJI"],
        scaled=True,
        meta=NDMeta({"exposure time": exposure_times}, axes={"exposure time": 0}, data_shape=data_1d.shape),
    )
    cube_1d.extra_coords.add(*EXTRA_COORDS[0])
    return cube_1d


@pytest.fixture
def dust_cube():
    data_dust = np.array(
        [
            [[-1, 2, -3, 4], [2, -200, 5, 3], [0, 1, 2, -300]],
            [[2, -200, 5, 1], [10, -5, 2, 2], [10, -3, 3, 0]],
        ],
    )
    header = {
        "CTYPE1": "HPLN-TAN",
        "CUNIT1": "arcsec",
        "CDELT1": 0.4,
        "CRPIX1": 0,
        "CRVAL1": 0,
        "NAXIS1": 4,
        "CTYPE2": "HPLT-TAN",
        "CUNIT2": "arcsec",
        "CDELT2": 0.5,
        "CRPIX2": 0,
        "CRVAL2": 0,
        "NAXIS2": 3,
        "CTYPE3": "Time    ",
        "CUNIT3": "s",
        "CDELT3": 0.3,
        "CRPIX3": 0,
        "CRVAL3": 0,
        "NAXIS3": 2,
    }
    wcs = WCS(header=header, naxis=3)
    unit = utils.DN_UNIT["SJI"]
    mask_dust = data_dust == -200

    uncertainty = 1
    times = Time(["2014-12-11T19:39:00.48", "2014-12-11T19:43:07.6"])
    exposure_times = 2 * np.ones((2), float) * u.s
    extra_coords = [("time", 0, times)]
    scaled_T = True
    meta = NDMeta(
        {"exposure time": exposure_times, "OBSID": 1},
        axes={"exposure time": 0},
        data_shape=data_dust.shape,
    )
    dust_cube = SJICube(
        data_dust,
        wcs,
        uncertainty=uncertainty,
        mask=mask_dust,
        unit=unit,
        scaled=scaled_T,
        meta=meta,
    )
    dust_cube.extra_coords.add(*extra_coords[0])
    return dust_cube


def test_sjicube_apply_dust_mask(dust_cube):
    # TODO: The expected values are not correct.
    dust_mask_expected = np.array(
        [
            [[True, True, True, True], [True, True, True, True], [True, True, False, False]],
            [[True, True, True, False], [True, True, True, True], [True, True, True, True]],
        ]
    )
    dust_cube.apply_dust_mask()
    np.testing.assert_array_equal(dust_cube.mask, dust_mask_expected)
    dust_cube.apply_dust_mask(undo=True)
    before_mask = np.array(
        [
            [[False, False, False, False], [False, False, False, False], [False, False, False, False]],
            [[False, False, False, False], [False, False, False, False], [False, False, False, False]],
        ]
    )
    np.testing.assert_array_equal(dust_cube.mask, before_mask)



"""
===============================
Working with spectrograph files
===============================

In this example, we will showcase how to open, crop and plot IRIS spectrograph data.
"""

import matplotlib.pyplot as plt
import numpy as np
import pooch

import astropy.units as u
from astropy.coordinates import SkyCoord, SpectralCoord
from astropy.visualization import quantity_support
from astropy.wcs.utils import wcs_to_celestial_frame

from sunpy.coordinates.frames import Helioprojective

from irispy.io import read_files

quantity_support()

###############################################################################
# We start with getting data from the IRIS archive.
#
# In this case, we will use ``pooch`` to keep this example self-contained
# but using your browser will also work.

raster_filename = pooch.retrieve(
    "http://www.lmsal.com/solarsoft/irisa/data/level2_compressed/2018/01/02/20180102_153155_3610108077/iris_l2_20180102_153155_3610108077_raster.tar.gz",
    known_hash="0ec2b7b20757c52b02e0d92c27a5852b6e28759512c3d455f8b6505d4e1f5cd6",
)

###############################################################################
# Note that when ``memmap=True``, the data values are read from the FITS file
# directly without the scaling to Float32, the data values are no longer in DN,
# but in scaled integer units that start at -2$^{16}$/2.

raster = read_files(raster_filename, memmap=False)

###############################################################################
# Let us now explore what was returned.

# Provides an overview of the Spectrograph object
print(raster)

# Will give us all the keys that corresponds to all the wavelength windows.
print(raster.keys())

# We can get the Mg II k window
mg_ii = raster["Mg II k 2796"]
print(mg_ii)

###############################################################################
# This is a `irispy.spectrograph.SpectrogramCubeSequence` which contains each
# complete raster as one individual `irispy.spectrograph.SpectrogramCube` object.
# In this case, it was only one complete raster, so the first axis is only length 1.
#
# So we will index to get the first raster and work with that.

mg_ii = mg_ii[0]
print(mg_ii)

###############################################################################
# Now we have more information about the data, including the OBS ID and description.
#
# Let's plot it

fig = plt.figure()
mg_ii.plot(fig=fig)

###############################################################################
# If we want to "raster" over wavelength, we can do the following

fig = plt.figure()
# This will also "transpose" the data but this is only for visualization purposes
# We have to set the vmin and vmax, as by default "plot" works out the
# vmin,vmax from the first slice which in this case is 0.
mg_ii.plot(fig=fig, plot_axes=["x", "y", None], vmin=0, vmax=1000).get_animation()
# The call to `get_animation()` is necessary to display the animation in this example.
# It is not needed in a normal script.

###############################################################################
# This object is sliceable, so we can do things like this:

print(mg_ii[120, 200])

fig = plt.figure()
ax = fig.add_subplot(111, projection=mg_ii[120, 200].wcs)
# This is just the data values along the wavelength axis of the Mg II k window at pixel (120, 200)
mg_ii[120, 200].plot(axes=ax)

###############################################################################
# We can also plot using the data directly. We can read the wavelengths of the
# Mg window by calling `ndcube.NDCube.axis_world_coords` for "wl" (wavelength), and redo the plot.

(mg_wave,) = mg_ii.axis_world_coords("wl")

fig, ax = plt.subplots()
ax.plot(mg_wave.to("AA"), mg_ii.data[120, 200])

###############################################################################
# When we use the underlying data directly, we lose all the metadata and WCS information.
#
# If you are unfamiliar with WCS, the following links are quite useful:
#
# * https://docs.astropy.org/en/stable/wcs/index.html
# * https://docs.astropy.org/en/stable/visualization/wcsaxes/index.html
#
# Some of the higher-level utilities are via ndcube, e.g., coordinate transformations: https://docs.sunpy.org/projects/ndcube/en/stable/explaining_ndcube/coordinates.html.
#
# Now, let's take a look at the WCS information.
# For example, what is the wavelength position that corresponds to Mg II k core (279.63 nm)?

iris_observer = wcs_to_celestial_frame(mg_ii.wcs.celestial).observer
iris_frame = Helioprojective(observer=iris_observer)
wcs_loc = mg_ii.wcs.world_to_pixel(
    SpectralCoord(279.63, unit=u.nm),
    SkyCoord(0 * u.arcsec, 0 * u.arcsec, frame=iris_frame),
)
mg_index = int(np.round(wcs_loc[0]))
print(mg_index)

###############################################################################
# Now we will plot spectroheliogram for Mg II k core wavelength.
# We can use the ``crop`` method to get this information, this will
# require a `~.SpectralCoord` object from `astropy.coordinates`.

# None, means that the axis is not cropped
# Note that this has to be in axis order
lower_corner = [SpectralCoord(280, unit=u.nm), None]
upper_corner = [SpectralCoord(280, unit=u.nm), None]
mg_spec_crop = mg_ii.crop(lower_corner, upper_corner)

fig = plt.figure()
ax = fig.add_subplot(111, projection=mg_spec_crop.wcs)
mg_spec_crop.plot(axes=ax)

###############################################################################
# Imagine there's a really cool feature at (-338", 275"), how can you plot
# the spectrum at that location?

lower_corner = [None, SkyCoord(-338 * u.arcsec, 275 * u.arcsec, frame=iris_frame)]
upper_corner = [None, SkyCoord(-338 * u.arcsec, 275 * u.arcsec, frame=iris_frame)]
mg_ii_cut = mg_ii.crop(lower_corner, upper_corner)

fig = plt.figure()
ax = fig.add_subplot(111, projection=mg_ii_cut.wcs)
mg_ii_cut.plot(axes=ax)

plt.show()

###############################################################################
#  Now, you may also be interested in knowing the time that was this observation taken.
# There is some information in ``.meta``.

print(mg_ii.meta)

###############################################################################
# But this is mostly about the observation in general.
# Times of individual scans are saved in .extra_coords['time'].
# Getting access to it can be done in the following  way:

print(mg_ii.axis_world_coords("time", wcs=mg_ii.extra_coords))



"""
==================
Mg II Dopplergrams
==================

In this example we are going to produce a Dopplergram for the Mg II k line from a
400-step raster. The Dopplergram is obtained by subtracting the intensities at
symmetrical velocity shifts from the line core (e.g., ±50 km/s).
"""

import tarfile

import matplotlib.pyplot as plt
import numpy as np
import pooch
from scipy.interpolate import interp1d

import astropy.units as u
from astropy import constants
from astropy.coordinates import SpectralCoord
from astropy.io import fits

from irispy.io import read_files
from irispy.utils import image_clipping

###############################################################################
# We start by getting the data from the IRIS archive.
#
# In this case, we will use ``pooch`` to keep this example self-contained
# but using your browser will also work.
#
# Using the url: http://www.lmsal.com/solarsoft/irisa/data/level2_compressed/2014/07/08/
# we are after `iris_l2_20140708_114109_3824262996_raster.tar.gz` which is ~730 MB in size.

downloaded_tar_iris_file = pooch.retrieve(
    "http://www.lmsal.com/solarsoft/irisa/data/level2_compressed/2014/07/08/20140708_114109_3824262996/iris_l2_20140708_114109_3824262996_raster.tar.gz",
    known_hash="0935729fad679bbe35f721b27040d808b7ed600d1a33d849656461388761fb4d",
)

###############################################################################
# Now to open the file using ``irispy-lmsal``.
# Note that when ``memmap=True``, the data values are read from the FITS file
# directly without the scaling to Float 32, the data values are no longer in DN,
# but in scaled integer units that start at -2$^{16}$/2.

raster = read_files(downloaded_tar_iris_file, memmap=True)

###############################################################################
# We are after the Mg II k window, which we can select using a key.

mg_ii = raster["Mg II k 2796"][0]
(mg_wave,) = mg_ii.axis_world_coords("wl")

# We will plot the spatially averaged spectrum
plt.figure()
plt.plot(mg_wave.to("nm"), mg_ii.data.mean((0, 1)))
plt.ylabel("DN (Memory Mapped Value)")
plt.xlabel("Wavelength (nm)")

###############################################################################
# This very large dense raster took more than three hours to complete
# across the 400 scans (with 30 s exposures), which means that the
# spacecraft's orbital velocity changes during the observations.
# This means that any calibration will need to correct for those shifts.
#
# To better understand the orbital velocity problem, let us look at how the
# line intensity varies for a strong Mn I line at around 280.2 nm, in
# between the Mg II k and h lines.
#
# For this dataset, the line core of this line falls around index 350.
# Here though, we will crop in wavelength space.

lower_corner = [SpectralCoord(280.2, unit=u.nm), None]
upper_corner = [SpectralCoord(280.2, unit=u.nm), None]
mg_crop = mg_ii.crop(lower_corner, upper_corner)
# We will "crunch" the image a bit
mg_crop.plot(aspect="auto")

###############################################################################
# You can see a regular bright-dark pattern along the x-axis, an
# indication that the intensities are not taken at the same position in
# the line because of wavelength shifts. The shifts are caused by the
# orbital velocity changes, and we can find these in the auxiliary
# metadata which are to be found in the extension past the "last" window
# in the FITS file.

# astropy.io.fits does not support opening tar files, so we need to extract.
with tarfile.open(downloaded_tar_iris_file, "r:gz") as tar_iris_file:
    tar_iris_file.extractall("./", filter="data")

# I know ahead of time what the filename is.
raster_filename = "iris_l2_20140708_114109_3824262996_raster_t000_r00000.fits"

# In this case, it is the 9th HDU, which we access directly
aux_data = fits.getdata(raster_filename, 9)
aux_header = fits.getheader(raster_filename, 9)
v_obs = aux_data[:, aux_header["OBS_VRIX"]] * u.m / u.s
# Convert to km/s as the data is in m/s
v_obs = v_obs.to("km/s")

plt.figure()
plt.plot(v_obs)
plt.ylabel("Orbital velocity (km/s)")
plt.xlabel("Scan number")

###############################################################################
# To look at intensities at any given scan we only need to subtract this
# velocity shift from the wavelength scale, but to look at the whole image
# at a given wavelength we must interpolate the original data to take this
# shift into account. Here is a way to do it (note that array dimensions
# apply to this specific example).

c = constants.c.to("km/s")
wave_shift = -v_obs * mg_wave[350] / c
# Linear interpolation in wavelength, for each scan
for i in range(mg_ii.data.shape[0]):
    mg_ii.data[:, i, :] = interp1d(mg_wave - wave_shift[i], mg_ii.data[:, i, :], bounds_error=False)(mg_wave)

###############################################################################
# Now we can plot the shifted data to see that the large scale shifts
# have disappeared

plt.figure()
# Since we changed the underlying data, we need to re-crop
mg_crop = mg_ii.crop(lower_corner, upper_corner)
# We will "crunch" the image a bit
mg_crop.plot(aspect="auto")

###############################################################################
# Some residual shift remains, but we will not correct for it here. A more
# elaborate correction can be obtained by the IDL routine
# ``iris_prep_wavecorr_l2``, but this has not yet been ported to Python
# see the `IDL version of this
# tutorial <http://iris.lmsal.com/itn26/tutorials.html#mg-ii-dopplergrams>`__
# for more details.
#
# We can use the calibrated data for example to calculate Dopplergrams. A
# Dopplergram is here defined as the difference between the intensities at
# two wavelength positions at the same (and opposite) distance from the
# line core. For example, at +/- 50 km/s from the Mg II k3 core. To do
# this, let us first calculate a velocity scale for the k line and find
# the indices of the -50 and +50 km/s velocity positions (here using the
# convention of negative velocities for up flows):

mg_k_centre = 279.6351 * u.nm
pos = 50 * u.km / u.s  # around line centre
velocity = (mg_wave - mg_k_centre) * c / mg_k_centre
index_p = np.argmin(np.abs(velocity - pos))
index_m = np.argmin(np.abs(velocity + pos))
doppler = mg_ii.data[..., index_m] - mg_ii.data[..., index_p]

###############################################################################
# And now we can plot this as before (intensity units are again arbitrary
# because of the unscaled DNs).

vmin, vmax = image_clipping(doppler)
plt.figure()
plt.imshow(
    doppler.T,
    cmap="RdBu",
    origin="lower",
    aspect=0.5,
    vmin=vmin,
    vmax=vmax,
)
plt.colorbar()
plt.xlabel("Solar X (arcsec)")
plt.ylabel("Solar Y (arcsec)")
plt.tight_layout()

plt.show()



"""
This module provides general utility functions.
"""

import numbers

import numpy as np
from scipy import interpolate, ndimage

import astropy.units as u
from astropy.modeling.models import custom_model

from irispy.utils.response import get_iris_response

__all__ = [
    "calculate_dust_mask",
    "calculate_uncertainty",
    "gaussian1d_on_linear_bg",
    "get_detector_type",
    "get_interpolated_effective_area",
    "image_clipping",
]


def image_clipping(image, cutoff=1.5e-3, gamma=1.0):
    """
    Computes and returns the min and max values of the input (image), clipping
    brightest and darkest pixels.

    Parameters
    ----------
    image : `numpy.ndarray`
        The input image.
    cutoff : float, optional
        The cutoff value for the histogram.
        Defaults to 1.5e-3
    gamma : float, optional
        The gamma value for the histogram.
        Defaults to 1.0

    References
    ----------
    Based on original IDL routine by P.Suetterlin (06 Jul 1993)
    Ported by V.Hansteen (15 Apr 2020)
    """
    hmin = np.min(image)
    hmax = np.max(image)
    if issubclass(image.dtype.type, numbers.Integral):
        nbins = np.abs(np.max(image) - np.min(image))
        hist = np.histogram(image, bins=nbins)
        fak = 1
    else:
        nbins = 10000
        fak = nbins / (hmax - hmin)
        hist = np.histogram((image - hmin) * fak, range=(0.0, float(nbins)), bins=nbins)
    h = hist[0]
    bins = hist[1]
    nh = np.size(h)
    # Integrate the histogram so that h(i) holds the number of points
    # with equal or lower intensity.
    for i in range(1, nh - 1):
        h[i] = h[i] + h[i - 1]
    h = h / float(h[nh - 2])
    h[nh - 1] = 1
    # As cutoff is in percent and h is normalized to unity,
    # vmin/vmax are the indices of the point where the number of pixels
    # with lower/higher intensity reach the given limit. This has to be
    # converted to a real image value by dividing by the scalefactor
    # fak and adding the min value of the image
    # Note that the bottom value is taken off (addition of h[0] to cutoff),
    # there are often very many points in IRIS images that are set to zero, this
    # removes them from calculation... and seems to work.
    vmin = (np.max(np.where(h <= (cutoff + h[0]), bins[1:] - bins[0], 0)) / fak + hmin) ** gamma
    vmax = (np.min(np.where(h >= (1.0 - cutoff), bins[1:] - bins[0], nh - 2)) / fak + hmin) ** gamma
    return vmin, vmax


@custom_model
def gaussian1d_on_linear_bg(
    x,
    amplitude=None,
    mean=None,
    standard_deviation=None,
    constant_term=None,
    linear_term=None,
):
    return amplitude * np.exp(-(((x - mean) / standard_deviation) ** 2)) + constant_term + linear_term * x


def get_detector_type(meta):
    """
    Gets the IRIS detector type from a meta dictionary.

    In this function, FUV1 and FUV2 are just assigned as FUV.

    Parameters
    ----------
    meta: dict-like
        Dictionary-like object containing entry for "detector type"

    Returns
    -------
    `str`
        Detector type.
    """
    return "FUV" if "FUV" in meta["detector type"] else meta["detector type"]


def get_interpolated_effective_area(time_obs, response_version, detector_type, obs_wavelength):
    """
    To compute the interpolated time-dependent effective area.

    It will generalize to the time of the observation.

    Parameters
    ----------
    time_obs : an `astropy.time.Time` object, as a kwarg, valid for version > 2
        Observation times of the datapoints.
        This argument is ignored for versions 1 and 2.
    response_version : `int`
        Version number of effective area file to be used. Cannot be set
        simultaneously with response_file or pre_launch kwarg. Default=4.
    detector_type : `str`
        Detector type: 'FUV' or 'NUV'.
    obs_wavelength : `astropy.units.Quantity`
        The wavelength at which the observation has been taken in Angstroms.

    Returns
    -------
    `numpy.array`
        The effective area(s) determined by interpolation with a spline fit.
    """
    iris_response = get_iris_response(time_obs, response_version=response_version)
    if detector_type == "FUV":
        detector_type_index = 0
    elif detector_type == "NUV":
        detector_type_index = 1
    else:
        msg = "Detector type not recognized."
        raise ValueError(msg)
    eff_area = iris_response["AREA_SG"][detector_type_index, :]
    response_wavelength = iris_response["LAMBDA"]
    # Interpolate the effective areas to cover the wavelengths
    # at which the data is recorded:
    eff_area_interp_base_unit = u.Angstrom
    tck = interpolate.splrep(
        response_wavelength.to(eff_area_interp_base_unit).value,
        eff_area.to(eff_area_interp_base_unit**2).value,
        s=0,
    )
    return interpolate.splev(obs_wavelength.to(eff_area_interp_base_unit).value, tck) * eff_area_interp_base_unit**2


def calculate_dust_mask(data_array):
    """
    Calculate a mask with the dust positions in a given array.

    Parameters
    ----------
    data_array : `numpy.ndarray`
        This array contains some dust position that will be calculated. The array
        must have scaled values.

    Returns
    -------
    `numpy.ndarray` of `bool`
        This array has the same shape than data_array and contains the dust positions
        when the value is True.
    """
    # Creating a mask with the same shape than the inputted data array.
    mask = np.zeros_like(data_array, dtype=bool)
    # Set the pixel value to True is the pixel is recognized as a dust pixel.
    mask[(data_array < 0.5) & (data_array > -200)] = True
    # Extending the mask to avoid the neighbours pixel influenced by the dust pixels.
    struct = np.array([np.zeros((3, 3)), np.ones((3, 3)), np.zeros((3, 3))], dtype=bool)
    return ndimage.binary_dilation(mask, structure=struct).astype(mask.dtype)


def calculate_uncertainty(data: np.array, readout_noise: u.Quantity, unit: u.Quantity) -> float:
    """
    Calculates the uncertainty of a given data array.

    Parameters
    ----------
    data : np.array
        The data array.
    readout_noise : u.Quantity
        The readout noise, needs to be a unit that is convertible to photon.
    unit : u.Quantity
        The final unit that the value should be converted to.

    Returns
    -------
    float
        The readout noise with no unit.
    """
    return (
        u.Quantity(
            np.sqrt((data * unit).to(u.photon).value + readout_noise.to(u.photon).value ** 2),
            unit=u.photon,
        )
        .to(unit)
        .value
    )
